@@ -5,16 +5,18 @@ import tempfile
 import numpy as np
 
 from nptdms import types
+from nptdms.timestamp import TimestampArray
 from nptdms.log import log_manager
 
 log = log_manager.get_logger(__name__)
 
 
-def get_data_receiver(obj, num_values, memmap_dir=None):
+def get_data_receiver(obj, num_values, raw_timestamps, memmap_dir=None):
     """Return a new channel data receiver to use for the given TDMS object
 
     :param obj: TDMS channel object to receive data for
     :param num_values: Number of values to be stored
+    :param raw_timestamps: Whether to store timestamp data as raw TDMS timestamps or a numpy datetime64 array
     :param memmap_dir: Optional directory to store memory map files,
         or None to not use memory map files
     """
@@ -24,6 +26,9 @@ def get_data_receiver(obj, num_values, memmap_dir=None):
     if obj.data_type == types.DaqMxRawData:
         return DaqmxDataReceiver(obj, num_values, memmap_dir)
 
+    if obj.data_type == types.TimeStamp:
+        return TimestampDataReceiver(obj, num_values, raw_timestamps, memmap_dir)
+
     if obj.data_type.nptype is None:
         return ListDataReceiver(obj)
 
@@ -124,6 +129,65 @@ class DaqmxDataReceiver(object):
         self._scaler_insert_positions[scale_id] += len(new_data)
 
 
+class TimestampDataReceiver(object):
+    """Receives timestamp data for a TDMS object and stores it in a numpy array
+
+    :ivar data: Data that has been read for the object as either a TimestampArray
+        or datetime64 array depending on whether raw_timestamps is True or False
+    """
+
+    def __init__(self, obj, num_values, raw_timestamps=False, memmap_dir=None):
+        """Initialise timestamp data receiver backed by a numpy array
+
+        :param obj: Object to store data for
+        :param num_values: Number of values to be stored
+        :param raw_timestamps: Whether to store data as raw TDMS timestamps or a numpy datetime64 array
+        :param memmap_dir: Optional directory to store memory map files in.
+        """
+
+        self.path = obj.path
+        self._raw_timestamps = raw_timestamps
+        if raw_timestamps:
+            byte_array = _new_numpy_array(np.dtype('uint8'), num_values * 16, memmap_dir)
+            dtype = np.dtype([('second_fractions', 'uint64'), ('seconds', 'int64')])
+            self.data = TimestampArray(byte_array.view(dtype))
+        else:
+            self.data = _new_numpy_array(np.dtype('datetime64[us]'), num_values, memmap_dir)
+        self.scaler_data = {}
+        self._data_insert_position = 0
+        log.debug("Allocated %d sample slots for %s", len(self.data), obj.path)
+
+    def append_data(self, new_data):
+        """Update the object data with a new array of data"""
+
+        log.debug("Adding %d data points to data for %s", len(new_data), self.path)
+        start_pos = self._data_insert_position
+        end_pos = self._data_insert_position + len(new_data)
+        if self._raw_timestamps:
+            # Need to be careful about potential endianness mismatch, so order of fields can differ
+            self.data['seconds'][start_pos:end_pos] = new_data['seconds']
+            self.data['second_fractions'][start_pos:end_pos] = new_data['second_fractions']
+        else:
+            self.data[start_pos:end_pos] = new_data.as_datetime64()
+        self._data_insert_position += len(new_data)
+
+
+class RawDataSlice(object):
+    def __init__(self, data, scaler_data):
+        self.data = data
+        self.scaler_data = scaler_data
+
+
+def slice_raw_data(raw_data, offset, length=None):
+    if offset == 0 and length is None:
+        return raw_data
+    end = None if length is None else offset + length
+    data = None if raw_data.data is None else raw_data.data[offset:end]
+    scaler_data = dict(
+        (scale_id, scaler_data[offset:end]) for (scale_id, scaler_data) in raw_data.scaler_data.items())
+    return RawDataSlice(data, scaler_data)
+
+
 def _new_numpy_array(dtype, num_values, memmap_dir=None):
     """Initialise a new numpy array for data
 
