@@ -4,17 +4,18 @@
 import logging
 import os
 import numpy as np
+import struct
 
 from nptdms import types
 from nptdms.common import ObjectPath, toc_properties
 from nptdms.utils import Timer, OrderedDict
 from nptdms.base_segment import RawChannelDataChunk
-from nptdms.tdms_segment import ContiguousDataSegment, InterleavedDataSegment
-from nptdms.daqmx import DaqmxSegment
+from nptdms.tdms_segment import TdmsSegment, SegmentIndexCache
 from nptdms.log import log_manager
 
 
 log = log_manager.get_logger(__name__)
+_struct_unpack = struct.unpack
 
 
 class TdmsReader(object):
@@ -34,9 +35,7 @@ class TdmsReader(object):
         self.object_metadata = OrderedDict()
         self._file_path = None
         self._index_file_path = None
-
-        self._segment_channel_offsets = None
-        self._segment_chunk_sizes = None
+        self._segment_channel_offsets = {}
 
         if hasattr(tdms_file, "read"):
             # Is a file
@@ -50,6 +49,10 @@ class TdmsReader(object):
                 self._index_file_path = index_file_path
 
     def close(self):
+        if self._file is None:
+            # Already closed
+            return
+
         if self._file_path is not None:
             # File path was provided so we opened the file and
             # should close it.
@@ -57,9 +60,12 @@ class TdmsReader(object):
         # Otherwise always remove reference to the file
         self._file = None
 
-    def read_metadata(self):
+    def read_metadata(self, require_segment_indexes=False):
         """ Read all metadata and structure information from a TdmsFile
+
+        :param require_segment_indexes: Whether to create segment object indexes to allow lookup of objects by path.
         """
+        self._ensure_open()
 
         if self._index_file_path is not None:
             reading_index_file = True
@@ -74,22 +80,26 @@ class TdmsReader(object):
             with Timer(log, "Read metadata"):
                 # Read metadata first to work out how much space we need
                 previous_segment = None
+                index_cache = SegmentIndexCache() if require_segment_indexes else None
                 while True:
+                    start_position = file.tell()
                     try:
-                        segment = self._read_segment_metadata(
-                            file, segment_position, previous_segment, reading_index_file)
+                        segment, properties = self._read_segment_metadata(
+                            file, segment_position, index_cache, previous_segment, reading_index_file)
                     except EOFError:
                         # We've finished reading the file
                         break
 
                     self._update_object_metadata(segment)
-                    self._update_object_properties(segment)
+                    self._update_object_properties(properties)
                     self._segments.append(segment)
                     previous_segment = segment
 
                     segment_position = segment.next_segment_pos
-                    if not reading_index_file:
-                        file.seek(segment.next_segment_pos)
+                    if reading_index_file:
+                        file.seek(start_position + segment.data_position - segment.position, os.SEEK_SET)
+                    else:
+                        file.seek(segment.next_segment_pos, os.SEEK_SET)
         finally:
             if reading_index_file:
                 file.close()
@@ -99,10 +109,12 @@ class TdmsReader(object):
 
         :returns: A generator that yields RawDataChunk objects
         """
+        self._ensure_open()
         if self._segments is None:
             raise RuntimeError(
                 "Cannot read data unless metadata has first been read")
         for segment in self._segments:
+            self._verify_segment_start(segment)
             for chunk in segment.read_raw_data(self._file):
                 yield chunk
 
@@ -116,33 +128,44 @@ class TdmsReader(object):
             Fewer values will be returned if attempting to read beyond the end of the available data.
         :returns: A generator that yields RawChannelDataChunk objects
         """
+        self._ensure_open()
         if self._segments is None:
             raise RuntimeError("Cannot read data unless metadata has first been read")
 
-        if self._segment_channel_offsets is None:
-            with Timer(log, "Build data index"):
-                self._build_index()
-        segment_offsets = self._segment_channel_offsets[channel_path]
-        chunk_sizes = self._segment_chunk_sizes[channel_path]
+        try:
+            (first_segment, segment_offsets) = self._segment_channel_offsets[channel_path]
+        except KeyError:
+            with Timer(log, "Build data index for channel"):
+                self._build_index(channel_path)
+            (first_segment, segment_offsets) = self._segment_channel_offsets[channel_path]
 
         object_metadata = self.object_metadata[channel_path]
+        max_length_from_offset = object_metadata.num_values - offset
         if length is None:
-            length = object_metadata.num_values - offset
+            length = max_length_from_offset
+        else:
+            # Make sure we're not trying to read more data than is actually available
+            length = min(length, max_length_from_offset)
         end_index = offset + length
 
         # Binary search to find first and last segments to read
-        start_segment = np.searchsorted(segment_offsets, offset, side='right')
-        end_segment = np.searchsorted(segment_offsets, end_index, side='left')
+        start_segment = first_segment + np.searchsorted(segment_offsets, offset, side='right')
+        end_segment = first_segment + np.searchsorted(segment_offsets, end_index, side='left')
 
         segment_index = start_segment
+        values_read = 0
         for segment in self._segments[start_segment:end_segment + 1]:
+            self._verify_segment_start(segment)
             # By default, read all chunks in a segment
             chunk_offset = 0
             num_chunks = segment.num_chunks
-            chunk_size = chunk_sizes[segment_index]
-            segment_start_index = 0 if segment_index == 0 else segment_offsets[segment_index - 1]
+            segment_obj = segment.get_segment_object(channel_path)
+            chunk_size = 0 if (segment_obj is None or not segment_obj.has_data) else segment_obj.number_values
+            if chunk_size == 0:
+                continue
+            segment_start_index = (
+                0 if segment_index == first_segment else segment_offsets[segment_index - first_segment - 1])
             remaining_values_to_skip = 0
-            remaining_values_to_trim = 0
 
             # For the first and last segments, we may not need to read all chunks,
             # and may need to trim some data from the beginning or end of the chunk.
@@ -153,7 +176,7 @@ class TdmsReader(object):
                 num_chunks -= chunk_offset
             if segment_index == end_segment:
                 # Note: segment_index may be both start and end
-                segment_end_index = segment_offsets[segment_index]
+                segment_end_index = segment_offsets[segment_index - first_segment]
                 num_values_to_trim = segment_end_index - end_index
 
                 # Account for segments where the final chunk is truncated
@@ -164,12 +187,12 @@ class TdmsReader(object):
                     num_values_to_trim -= final_chunk_size
 
                 num_chunks -= num_values_to_trim // chunk_size
-                remaining_values_to_trim = num_values_to_trim % chunk_size
 
             for i, chunk in enumerate(
                     segment.read_raw_data_for_channel(self._file, channel_path, chunk_offset, num_chunks)):
                 skip = remaining_values_to_skip if i == 0 else 0
-                trim = remaining_values_to_trim if i == (num_chunks - 1) else 0
+                values_read += len(chunk) - skip
+                trim = 0 if values_read < length else values_read - length
                 yield _trim_channel_chunk(chunk, skip, trim)
 
             segment_index += 1
@@ -180,97 +203,105 @@ class TdmsReader(object):
         :returns: Tuple of raw channel data chunk and the integer offset to the beginning of the chunk
         :rtype: (RawChannelDataChunk, int)
         """
+        self._ensure_open()
         if self._segments is None:
             raise RuntimeError("Cannot read data unless metadata has first been read")
 
-        if self._segment_channel_offsets is None:
-            with Timer(log, "Build data index"):
-                self._build_index()
-        segment_offsets = self._segment_channel_offsets[channel_path]
+        try:
+            (first_segment, segment_offsets) = self._segment_channel_offsets[channel_path]
+        except KeyError:
+            with Timer(log, "Build data index for channel"):
+                self._build_index(channel_path)
+            (first_segment, segment_offsets) = self._segment_channel_offsets[channel_path]
 
         # Binary search to find the segment to read
-        segment_index = np.searchsorted(segment_offsets, index, side='right')
+        segment_index = first_segment + np.searchsorted(segment_offsets, index, side='right')
         segment = self._segments[segment_index]
-        chunk_size = self._segment_chunk_sizes[channel_path][segment_index]
-        segment_start_index = segment_offsets[segment_index - 1] if segment_index > 0 else 0
+        segment_obj = segment.get_segment_object(channel_path)
+        chunk_size = 0 if segment_obj is None else segment_obj.number_values
+        segment_start_index = (
+            0 if segment_index == first_segment else segment_offsets[segment_index - first_segment - 1])
 
         index_in_segment = index - segment_start_index
         chunk_index = index_in_segment // chunk_size
 
+        self._verify_segment_start(segment)
         chunk_data = next(segment.read_raw_data_for_channel(self._file, channel_path, chunk_index, 1))
         chunk_offset = segment_start_index + chunk_index * chunk_size
         return chunk_data, chunk_offset
 
     def _read_segment_metadata(
-            self, file, segment_position, previous_segment=None, is_index_file=False):
-        (position, toc_mask, endianness, data_position, raw_data_offset,
-         next_segment_offset, next_segment_pos) = self._read_lead_in(file, segment_position, is_index_file)
+            self, file, segment_position, index_cache, previous_segment, is_index_file):
+        (position, toc_mask, data_position, next_segment_pos, segment_incomplete) = self._read_lead_in(
+            file, segment_position, is_index_file)
 
-        segment_args = (
-            position, toc_mask, endianness, next_segment_offset,
-            next_segment_pos, raw_data_offset, data_position)
-        if toc_mask & toc_properties['kTocDAQmxRawData']:
-            segment = DaqmxSegment(*segment_args)
-        elif toc_mask & toc_properties['kTocInterleavedData']:
-            segment = InterleavedDataSegment(*segment_args)
-        else:
-            segment = ContiguousDataSegment(*segment_args)
+        segment = TdmsSegment(
+            position, toc_mask, next_segment_pos, data_position)
 
-        segment.read_segment_objects(
-            file, self._prev_segment_objects, previous_segment)
-        return segment
+        properties = segment.read_segment_objects(
+            file, self._prev_segment_objects, index_cache, previous_segment, segment_incomplete)
+        return segment, properties
 
     def _read_lead_in(self, file, segment_position, is_index_file=False):
+        lead_in_bytes = file.read(28)
+
         expected_tag = b'TDSh' if is_index_file else b'TDSm'
-        tag = file.read(4)
+        tag = lead_in_bytes[:4]
         if tag == b'':
             raise EOFError
         if tag != expected_tag:
             raise ValueError(
                 "Segment does not start with %r, but with %r" % (expected_tag, tag))
 
-        log.debug("Reading segment at %d", segment_position)
-
         # Next four bytes are table of contents mask
-        toc_mask = types.Int32.read(file)
+        toc_mask = _struct_unpack('<l', lead_in_bytes[4:8])[0]
 
         if log.isEnabledFor(logging.DEBUG):
+            log.debug("Reading segment at %d", segment_position)
             for prop_name, prop_mask in toc_properties.items():
                 prop_is_set = (toc_mask & prop_mask) != 0
                 log.debug("Property %s is %s", prop_name, prop_is_set)
 
         endianness = '>' if (toc_mask & toc_properties['kTocBigEndian']) else '<'
 
-        # Next four bytes are version number
-        version = types.Int32.read(file, endianness)
+        # Next four bytes are version number, then 8 bytes each for the offset values
+        (version, next_segment_offset, raw_data_offset) = _struct_unpack(endianness + 'lQQ', lead_in_bytes[8:28])
+
         if version not in (4712, 4713):
             log.warning("Unrecognised version number.")
 
-        # Now 8 bytes each for the offset values
-        next_segment_offset = types.Uint64.read(file, endianness)
-        raw_data_offset = types.Uint64.read(file, endianness)
-
         # Calculate data and next segment position
         lead_size = 7 * 4
         data_position = segment_position + lead_size + raw_data_offset
-        if next_segment_offset == 0xFFFFFFFFFFFFFFFF:
+        segment_incomplete = next_segment_offset == 0xFFFFFFFFFFFFFFFF
+        if segment_incomplete:
             # Segment size is unknown. This can happen if LabVIEW crashes.
             # Try to read until the end of the file.
             log.warning(
                 "Last segment of file has unknown size, "
                 "will attempt to read to the end of the file")
             next_segment_pos = self._get_data_file_size()
-            next_segment_offset = next_segment_pos - segment_position - lead_size
         else:
-            log.debug("Next segment offset = %d, raw data offset = %d",
-                      next_segment_offset, raw_data_offset)
-            log.debug("Data size = %d b",
-                      next_segment_offset - raw_data_offset)
+            log.debug("Next segment offset = %d, raw data offset = %d, data size = %d b",
+                      next_segment_offset, raw_data_offset, next_segment_offset - raw_data_offset)
             next_segment_pos = (
                     segment_position + next_segment_offset + lead_size)
 
-        return (segment_position, toc_mask, endianness, data_position, raw_data_offset,
-                next_segment_offset, next_segment_pos)
+        return segment_position, toc_mask, data_position, next_segment_pos, segment_incomplete
+
+    def _verify_segment_start(self, segment):
+        """ When reading data for a segment, check for the TDSm tag at the start of the segment in an attempt
+            to detect any mismatch between tdms and tdms_index files.
+        """
+        position = segment.position
+        self._file.seek(segment.position)
+        expected_tag = b'TDSm'
+        tag = self._file.read(4)
+        if tag != expected_tag:
+            raise ValueError(
+                "Attempted to read data segment at position {0} but did not find segment start header. ".format(
+                    position) +
+                "Check that the tdms_index file matches the tdms data file.")
 
     def _get_data_file_size(self):
         current_pos = self._file.tell()
@@ -289,13 +320,14 @@ class TdmsReader(object):
             object_metadata = self._get_or_create_object(path)
             object_metadata.num_values += _number_of_segment_values(segment_object, segment)
             _update_object_data_type(path, object_metadata, segment_object)
-            _update_object_scaler_data_types(path, object_metadata, segment_object)
+            if segment_object.scaler_data_types is not None:
+                _update_object_scaler_data_types(path, object_metadata, segment_object)
 
-    def _update_object_properties(self, segment):
+    def _update_object_properties(self, segment_object_properties):
         """ Update object properties using any properties in a segment
         """
-        if segment.object_properties is not None:
-            for path, properties in segment.object_properties.items():
+        if segment_object_properties is not None:
+            for path, properties in segment_object_properties.items():
                 object_metadata = self._get_or_create_object(path)
                 for prop, val in properties:
                     object_metadata.properties[prop] = val
@@ -310,33 +342,46 @@ class TdmsReader(object):
             self.object_metadata[path] = obj
             return obj
 
-    def _build_index(self):
+    def _build_index(self, channel_path):
         """ Builds an index into the segment data for faster lookup of values
 
             _segment_channel_offsets provides data offset at the end of each segment per channel
-            _segment_chunk_sizes provides chunk sizes in each segment per channel
         """
-        data_objects = [
-            path
-            for (path, obj) in self.object_metadata.items()
-            if ObjectPath.from_string(path).is_channel]
         num_segments = len(self._segments)
 
-        segment_num_values = {
-            path: np.zeros(num_segments, dtype=np.int64) for path in data_objects}
-        segment_chunk_sizes = {
-            path: np.zeros(num_segments, dtype=np.int64) for path in data_objects}
+        # Get number of values for this channel in each segment
+        segment_num_values = np.zeros(num_segments, dtype=np.int64)
+        first_segment = -1
+        last_segment = -1
 
         for i, segment in enumerate(self._segments):
-            for obj in segment.ordered_objects:
-                if not obj.has_data:
-                    continue
-                segment_chunk_sizes[obj.path][i] = obj.number_values if obj.has_data else 0
-                segment_num_values[obj.path][i] = _number_of_segment_values(obj, segment)
+            obj_index = segment.object_index.get(channel_path)
+            if obj_index is not None:
+                segment_obj = segment.ordered_objects[obj_index]
+                num_values = _number_of_segment_values(segment_obj, segment)
+                if num_values > 0:
+                    segment_num_values[i] = num_values
+                    last_segment = i
+                    if first_segment == -1:
+                        first_segment = i
 
-        self._segment_chunk_sizes = segment_chunk_sizes
-        self._segment_channel_offsets = {
-            path: np.cumsum(segment_count) for (path, segment_count) in segment_num_values.items()}
+        # Now use the cumulative sum to get the total channel value count
+        # at the end of each segment.
+        if first_segment == -1:
+            first_segment = num_segments
+            last_segment = num_segments
+        channel_offsets = np.cumsum(segment_num_values[first_segment:last_segment + 1])
+
+        # It's likely that many channels will have the same shaped data,
+        # so de-duplicate these arrays to reduce memory usage.
+        existing_arrays = (xs for (_, xs) in self._segment_channel_offsets.values())
+        channel_offsets = _deduplicate_array(channel_offsets, existing_arrays)
+        self._segment_channel_offsets[channel_path] = (first_segment, channel_offsets)
+
+    def _ensure_open(self):
+        if self._file is None:
+            raise RuntimeError(
+                "Cannot read data after the underlying TDMS reader is closed")
 
 
 def _number_of_segment_values(segment_object, segment):
@@ -344,13 +389,11 @@ def _number_of_segment_values(segment_object, segment):
     """
     if not segment_object.has_data:
         return 0
-    num_chunks = segment.num_chunks
-    final_chunk_proportion = segment.final_chunk_proportion
-    if final_chunk_proportion == 1.0:
-        return segment_object.number_values * num_chunks
+    if segment.final_chunk_lengths_override is None:
+        return segment_object.number_values * segment.num_chunks
     else:
-        return (segment_object.number_values * (num_chunks - 1) +
-                int(segment_object.number_values * final_chunk_proportion))
+        return (segment_object.number_values * (segment.num_chunks - 1) +
+                segment.final_chunk_lengths_override.get(segment_object.path, 0))
 
 
 def _update_object_data_type(path, obj, segment_object):
@@ -367,13 +410,12 @@ def _update_object_data_type(path, obj, segment_object):
 def _update_object_scaler_data_types(path, obj, segment_object):
     """ Update the DAQmx scaler data types for an object using its segment metadata
     """
-    if segment_object.scaler_data_types is not None:
-        if obj.scaler_data_types is not None and obj.scaler_data_types != segment_object.scaler_data_types:
-            raise ValueError(
-                "Segment data doesn't have the same scaler data types as previous "
-                "segments for objects %s. Expected types %s but got %s" %
-                (path, obj.scaler_data_types, segment_object.scaler_data_types))
-        obj.scaler_data_types = segment_object.scaler_data_types
+    if obj.scaler_data_types is not None and obj.scaler_data_types != segment_object.scaler_data_types:
+        raise ValueError(
+            "Segment data doesn't have the same scaler data types as previous "
+            "segments for objects %s. Expected types %s but got %s" %
+            (path, obj.scaler_data_types, segment_object.scaler_data_types))
+    obj.scaler_data_types = segment_object.scaler_data_types
 
 
 class ObjectMetadata(object):
@@ -398,3 +440,29 @@ def _trim_channel_chunk(chunk, skip=0, trim=0):
             scale_id: d[skip:len(d) - trim]
             for (scale_id, d) in chunk.scaler_data.items()}
     return RawChannelDataChunk(data, scaler_data)
+
+
+def _deduplicate_array(xs, candidates):
+    """ Reduce memory usage by replacing an array with a reference to an existing array if equal
+    """
+    for candidate in candidates:
+        if _array_equal(xs, candidate):
+            return candidate
+    return xs
+
+
+def _array_equal(a, b, chunk_size=100):
+    """ Compare two arrays for equality
+    """
+    # Numpy array_equal compares all elements rather than comparing one at a time and short-circuiting when it
+    # finds a difference. Break up the comparison into chunks to make this faster. Adapted from:
+    # https://stackoverflow.com/questions/26260848/numpy-fast-check-for-complete-array-equality-like-matlabs-isequal
+    if len(a) != len(b):
+        return False
+
+    num_chunks = (len(a) + chunk_size - 1) // chunk_size
+    for i in range(num_chunks):
+        offset = i * chunk_size
+        if not (a[offset:offset+chunk_size] == b[offset:offset+chunk_size]).all():
+            return False
+    return True
