@@ -1,7 +1,9 @@
 import numpy as np
+import numpy.polynomial.polynomial as poly
 import re
 
 from nptdms.log import log_manager
+import nptdms.thermocouples as thermocouples
 
 
 log = log_manager.get_logger(__name__)
@@ -11,6 +13,25 @@ VOLTAGE_EXCITATION = 10322
 CURRENT_EXCITATION = 10134
 
 
+class NoOpScaling(object):
+    """ Does not apply any scaling
+    """
+    def __init__(self, input_source):
+        self.input_source = input_source
+
+    @staticmethod
+    def from_properties(properties, scale_index, scale_name):
+        try:
+            input_source = properties[
+                "NI_Scale[%d]_%s_Input_Source" % (scale_index, scale_name)]
+        except KeyError:
+            input_source = RAW_DATA_INPUT_SOURCE
+        return NoOpScaling(input_source)
+
+    def scale(self, data):
+        return data
+
+
 class LinearScaling(object):
     """ Linear scaling with slope and intercept
     """
@@ -71,7 +92,7 @@ class PolynomialScaling(object):
 
 class RtdScaling(object):
     """ Converts a signal from a resistance temperature detector into
-        degrees celcius using the Callendar-Van Dusen equation
+        degrees Celsius using the Callendar-Van Dusen equation
     """
     def __init__(
             self, current_excitation, r0_nominal_resistance,
@@ -109,29 +130,89 @@ class RtdScaling(object):
     def scale(self, data):
         """ Convert voltage data to temperature
         """
-        r_0 = self.r0_nominal_resistance
-        a = self.a
-        b = self.b
+        (a, b, r_0) = (self.a, self.b, self.r0_nominal_resistance)
 
         # R(T) = R(0)[1 + A*T + B*T^2 + (T - 100)*C*T^3]
         # R(T) = V/I
 
         r_t = data / self.current_excitation
-
         r_t = _adjust_for_lead_resistance(
             r_t, CURRENT_EXCITATION, self.resistance_configuration, self.lead_wire_resistance)
 
-        if np.all(r_t >= self.r0_nominal_resistance):
-            # For R(T) > R(0), temperature is positive and we can use the
-            # quadratic form of the equation without C:
-            return ((-a + np.sqrt(a ** 2 - 4.0 * b * (1.0 - r_t / r_0))) /
-                    (2.0 * self.b))
-        else:
-            # We would need to solve for the roots of the full quartic equation
-            # for any cases where R(T) < R(0), and work out which root is the
-            # correct solution. For R(T) > R(0) we set C to zero.
-            raise NotImplementedError(
-                "RTD scaling for temperatures < 0 is not implemented")
+        positive_temperature = r_t >= r_0
+        # First solve for positive temperatures using the quadratic form
+        temperature = (-a + np.sqrt(a ** 2 - 4.0 * b * (1.0 - r_t / r_0), where=positive_temperature)) / (2.0 * b)
+        if not np.all(positive_temperature):
+            # Use full quartic for any negative temperatures
+            for i in np.where(np.logical_not(positive_temperature))[0]:
+                temperature[i] = self._solve_quartic_form(r_t[i])
+        return temperature
+
+    def _solve_quartic_form(self, r_t):
+        (a, b, c, r_0) = (self.a, self.b, self.c, self.r0_nominal_resistance)
+        poly_coefficients = [r_0 - r_t, r_0 * a, r_0 * b, -100.0 * r_0 * c, r_0 * c]
+        roots = poly.polyroots(poly_coefficients)
+        return RtdScaling._get_negative_real_root(roots)
+
+    @staticmethod
+    def _get_negative_real_root(roots):
+        filtered = [r for r in roots if not np.iscomplex(r) and r.real < 0.0]
+        if len(filtered) != 1:
+            raise ValueError("Expected single real valued negative root for RTD equation")
+        return filtered[0].real
+
+
+class StrainScaling(object):
+    """ Converts a voltage measurement from a strain gauge bridge to strain
+    """
+    def __init__(
+            self, configuration, poisson_ratio, gage_resistance, lead_wire_resistance, initial_bridge_voltage,
+            gage_factor, gain_adjustment, voltage_excitation, input_source):
+        self.configuration = configuration
+        self.poisson_ratio = poisson_ratio
+        self.gage_resistance = gage_resistance
+        self.lead_wire_resistance = lead_wire_resistance
+        self.initial_bridge_voltage = initial_bridge_voltage
+        self.gage_factor = gage_factor
+        self.gain_adjustment = gain_adjustment
+        self.voltage_excitation = voltage_excitation
+        self.input_source = input_source
+
+    @staticmethod
+    def from_properties(properties, scale_index):
+        prefix = "NI_Scale[%d]_Strain" % scale_index
+        configuration = properties["%s_Configuration" % prefix]
+        poisson_ratio = properties["%s_Poisson_Ratio" % prefix]
+        gage_resistance = properties["%s_Gage_Resistance" % prefix]
+        lead_wire_resistance = properties["%s_Lead_Wire_Resistance" % prefix]
+        initial_bridge_voltage = properties["%s_Initial_Bridge_Voltage" % prefix]
+        gage_factor = properties["%s_Gage_Factor" % prefix]
+        gain_adjustment = properties["%s_Bridge_Shunt_Calibration_Gain_Adjustment" % prefix]
+        voltage_excitation = properties["%s_Voltage_Excitation" % prefix]
+        input_source = properties["%s_Input_Source" % prefix]
+        return StrainScaling(
+            configuration, poisson_ratio, gage_resistance, lead_wire_resistance, initial_bridge_voltage, gage_factor,
+            gain_adjustment, voltage_excitation, input_source)
+
+    def scale(self, data):
+        """ Convert voltage data to strain
+        """
+        if self.configuration == StrainScaling.FULL_BRIDGE_1:
+            if self.initial_bridge_voltage != 0.0:
+                voltage_diff = data - self.initial_bridge_voltage
+            else:
+                voltage_diff = data
+            return voltage_diff * (-self.gain_adjustment / (self.voltage_excitation * self.gage_factor))
+
+        raise Exception("Strain gauge configuration %d is not supported" % self.configuration)
+
+    FULL_BRIDGE_1 = 10183
+    FULL_BRIDGE_2 = 10184
+    FULL_BRIDGE_3 = 10185
+    HALF_BRIDGE_1 = 10188
+    HALF_BRIDGE_2 = 10189
+    QUARTER_BRIDGE_1 = 10271
+    QUARTER_BRIDGE_2 = 10272
 
 
 class TableScaling(object):
@@ -253,26 +334,24 @@ class ThermistorScaling(object):
 
 
 class ThermocoupleScaling(object):
-    """ Convert between voltages in uV and degrees celcius for a Thermocouple.
+    """ Convert between voltages in uV and degrees Celsius for a Thermocouple.
         Can convert in either direction depending on the scaling direction
         parameter.
     """
     def __init__(self, type_code, scaling_direction, input_source):
-        from thermocouples_reference import thermocouples
-
         # Thermocouple types from
         # http://zone.ni.com/reference/en-XX/help/371361R-01/glang/tdms_create_scalinginfo/#instance2
-        thermocouple_type = {
-            10047: 'B',
-            10055: 'E',
-            10072: 'J',
-            10073: 'K',
-            10077: 'N',
-            10082: 'R',
-            10085: 'S',
-            10086: 'T',
-        }[type_code]
-        self.thermocouple = thermocouples[thermocouple_type]
+        thermocouple_types = {
+            10047: thermocouples.type_b,
+            10055: thermocouples.type_e,
+            10072: thermocouples.type_j,
+            10073: thermocouples.type_k,
+            10077: thermocouples.type_n,
+            10082: thermocouples.type_r,
+            10085: thermocouples.type_s,
+            10086: thermocouples.type_t,
+        }
+        self.thermocouple = thermocouple_types[type_code]
 
         self.scaling_direction = scaling_direction
         self.input_source = input_source
@@ -291,31 +370,12 @@ class ThermocoupleScaling(object):
     def scale(self, data):
         """ Apply thermocouple scaling
         """
-
-        # Note that the thermocouples_reference package uses mV for voltages,
-        # but TDMS uses uV.
-        nan = float('nan')
-
-        def scale_uv_to_c(micro_volts):
-            """Convert micro volts to degrees celcius"""
-            milli_volts = micro_volts / 1000.0
-            try:
-                return self.thermocouple.inverse_CmV(milli_volts, Tref=0.0)
-            except ValueError:
-                return nan
-
-        def scale_c_to_uv(temp):
-            """Convert degrees celcius to micro volts"""
-            try:
-                return 1000.0 * self.thermocouple.emf_mVC(temp, Tref=0.0)
-            except ValueError:
-                return nan
-
+        # Note that the thermocouple conversions use mV for voltages, but TDMS uses uV.
         if self.scaling_direction == 1:
-            scaled = np.vectorize(scale_c_to_uv)(data)
+            return 1000.0 * self.thermocouple.celsius_to_mv(data)
         else:
-            scaled = np.vectorize(scale_uv_to_c)(data)
-        return scaled
+            milli_volts = data / 1000.0
+            return self.thermocouple.mv_to_celsius(milli_volts)
 
 
 class AddScaling(object):
@@ -397,6 +457,8 @@ class MultiScaling(object):
             return np.result_type(
                 self._compute_scale_dtype(scaling.left_input_source, raw_data_type, scaler_data_types),
                 self._compute_scale_dtype(scaling.right_input_source, raw_data_type, scaler_data_types))
+        elif isinstance(scaling, NoOpScaling):
+            return raw_data_type.nptype
         else:
             # Any other scaling type should produce double data
             return np.dtype('float64')
@@ -472,6 +534,9 @@ def _get_channel_scaling(properties):
         elif scale_type == 'RTD':
             scalings[scale_index] = RtdScaling.from_properties(
                 properties, scale_index)
+        elif scale_type == 'Strain':
+            scalings[scale_index] = StrainScaling.from_properties(
+                properties, scale_index)
         elif scale_type == 'Table':
             scalings[scale_index] = TableScaling.from_properties(
                 properties, scale_index)
@@ -487,6 +552,9 @@ def _get_channel_scaling(properties):
         elif scale_type == 'Subtract':
             scalings[scale_index] = SubtractScaling.from_properties(
                 properties, scale_index)
+        elif scale_type == 'AdvancedAPI':
+            scalings[scale_index] = NoOpScaling.from_properties(
+                properties, scale_index, 'AdvancedAPI')
         else:
             log.warning("Unsupported scale type: %s", scale_type)
             return None
