@@ -4,7 +4,6 @@
 """
 
 from collections import defaultdict
-import warnings
 import numpy as np
 
 from nptdms import scaling, types
@@ -12,19 +11,15 @@ from nptdms.utils import Timer, OrderedDict, cached_property
 from nptdms.log import log_manager
 from nptdms.common import ObjectPath
 from nptdms.reader import TdmsReader
-from nptdms.channel_data import get_data_receiver
+from nptdms.channel_data import get_data_receiver, slice_raw_data
 from nptdms.export import hdf_export, pandas_export
 from nptdms.base_segment import RawChannelDataChunk
+from nptdms.timestamp import TdmsTimestamp, TimestampArray
 
 
 log = log_manager.get_logger(__name__)
 
 
-# Have to get a reference to the builtin property decorator
-# so we can use it in TdmsObject, which has a property method.
-_property_builtin = property
-
-
 class TdmsFile(object):
     """ Reads and stores data from a TDMS file.
 
@@ -57,48 +52,61 @@ class TdmsFile(object):
     """
 
     @staticmethod
-    def read(file, memmap_dir=None):
+    def read(file, raw_timestamps=False, memmap_dir=None):
         """ Creates a new TdmsFile object and reads all data in the file
 
         :param file: Either the path to the tdms file to read
             as a string or pathlib.Path, or an already opened file.
+        :param raw_timestamps: By default TDMS timestamps are read as numpy datetime64
+            but this loses some precision.
+            Setting this to true will read timestamps as a custom TdmsTimestamp type.
         :param memmap_dir: The directory to store memory mapped data files in,
             or None to read data into memory. The data files are created
             as temporary files and are deleted when the channel data is no
             longer used. tempfile.gettempdir() can be used to get the default
             temporary file directory.
         """
-        return TdmsFile(file, memmap_dir=memmap_dir)
+        return TdmsFile(file, raw_timestamps=raw_timestamps, memmap_dir=memmap_dir)
 
     @staticmethod
-    def open(file, memmap_dir=None):
+    def open(file, raw_timestamps=False, memmap_dir=None):
         """ Creates a new TdmsFile object and reads metadata, leaving the file open
             to allow reading channel data
 
         :param file: Either the path to the tdms file to read
             as a string or pathlib.Path, or an already opened file.
+        :param raw_timestamps: By default TDMS timestamps are read as numpy datetime64
+            but this loses some precision.
+            Setting this to true will read timestamps as a custom TdmsTimestamp type.
         :param memmap_dir: The directory to store memory mapped data files in,
             or None to read data into memory. The data files are created
             as temporary files and are deleted when the channel data is no
             longer used. tempfile.gettempdir() can be used to get the default
             temporary file directory.
         """
-        return TdmsFile(file, memmap_dir=memmap_dir, read_metadata_only=True, keep_open=True)
+        return TdmsFile(
+            file, raw_timestamps=raw_timestamps, memmap_dir=memmap_dir, read_metadata_only=True, keep_open=True)
 
     @staticmethod
-    def read_metadata(file):
+    def read_metadata(file, raw_timestamps=False):
         """ Creates a new TdmsFile object and only reads the metadata
 
         :param file: Either the path to the tdms file to read
             as a string or pathlib.Path, or an already opened file.
+        :param raw_timestamps: By default TDMS timestamps are read as numpy datetime64
+            but this loses some precision.
+            Setting this to true will read timestamps as a custom TdmsTimestamp type.
         """
-        return TdmsFile(file, read_metadata_only=True)
+        return TdmsFile(file, raw_timestamps=raw_timestamps, read_metadata_only=True)
 
-    def __init__(self, file, memmap_dir=None, read_metadata_only=False, keep_open=False):
+    def __init__(self, file, raw_timestamps=False, memmap_dir=None, read_metadata_only=False, keep_open=False):
         """Initialise a new TdmsFile object
 
         :param file: Either the path to the tdms file to read
             as a string or pathlib.Path, or an already opened file.
+        :param raw_timestamps: By default TDMS timestamps are read as numpy datetime64
+            but this loses some precision.
+            Setting this to true will read timestamps as a custom TdmsTimestamp type.
         :param memmap_dir: The directory to store memory mapped data files in,
             or None to read data into memory. The data files are created
             as temporary files and are deleted when the channel data is no
@@ -111,20 +119,18 @@ class TdmsFile(object):
         """
 
         self._memmap_dir = memmap_dir
+        self._raw_timestamps = raw_timestamps
         self._groups = OrderedDict()
-        self._properties = {}
+        self._properties = OrderedDict()
         self._channel_data = {}
-        self._reader = None
         self.data_read = False
 
-        reader = TdmsReader(file)
+        self._reader = TdmsReader(file)
         try:
-            self._read_file(reader, read_metadata_only)
+            self._read_file(self._reader, read_metadata_only, keep_open)
         finally:
-            if keep_open:
-                self._reader = reader
-            else:
-                reader.close()
+            if not keep_open:
+                self._reader.close()
 
     def groups(self):
         """Returns a list of the groups in this file
@@ -134,7 +140,7 @@ class TdmsFile(object):
 
         return list(self._groups.values())
 
-    @_property_builtin
+    @property
     def properties(self):
         """ Return the properties of this file as a dictionary
 
@@ -175,9 +181,9 @@ class TdmsFile(object):
 
         :rtype: Generator that yields :class:`DataChunk` objects
         """
-        reader = self._get_reader()
         channel_offsets = defaultdict(int)
-        for chunk in reader.read_raw_data():
+        for chunk in self._reader.read_raw_data():
+            _convert_data_chunk(chunk, self._raw_timestamps)
             yield DataChunk(self, chunk, channel_offsets)
             for path, data in chunk.channel_data.items():
                 channel_offsets[path] += len(data)
@@ -216,30 +222,37 @@ class TdmsFile(object):
     def __exit__(self, exc_type, exc_value, traceback):
         self.close()
 
-    def _get_reader(self):
-        if self._reader is None:
-            raise RuntimeError(
-                "Cannot read data after the underlying TDMS reader is closed")
-        return self._reader
-
-    def _read_file(self, tdms_reader, read_metadata_only):
-        tdms_reader.read_metadata()
+    def _read_file(self, tdms_reader, read_metadata_only, keep_open):
+        tdms_reader.read_metadata(require_segment_indexes=keep_open)
 
         # Use object metadata to build group and channel objects
         group_properties = OrderedDict()
         group_channels = OrderedDict()
+        object_properties = {
+            path_string: self._convert_properties(obj.properties)
+            for path_string, obj in tdms_reader.object_metadata.items()}
+        try:
+            self._properties = object_properties['/']
+        except KeyError:
+            pass
+
         for (path_string, obj) in tdms_reader.object_metadata.items():
+            properties = object_properties[path_string]
             path = ObjectPath.from_string(path_string)
             if path.is_root:
-                # Root object provides properties for the whole file
-                self._properties = obj.properties
+                pass
             elif path.is_group:
-                group_properties[path.group] = obj.properties
+                group_properties[path.group] = properties
             else:
                 # Object is a channel
+                try:
+                    channel_group_properties = object_properties[path.group_path()]
+                except KeyError:
+                    channel_group_properties = OrderedDict()
                 channel = TdmsChannel(
-                    self, path, obj.properties, obj.data_type,
-                    obj.scaler_data_types, obj.num_values)
+                    path, obj.data_type, obj.scaler_data_types, obj.num_values,
+                    properties, channel_group_properties, self._properties,
+                    tdms_reader, self._raw_timestamps, self._memmap_dir)
                 if path.group in group_channels:
                     group_channels[path.group].append(channel)
                 else:
@@ -268,7 +281,7 @@ class TdmsFile(object):
             for group in self.groups():
                 for channel in group.channels():
                     self._channel_data[channel.path] = get_data_receiver(
-                        channel, len(channel), self._memmap_dir)
+                        channel, len(channel), self._raw_timestamps, self._memmap_dir)
 
         with Timer(log, "Read data"):
             # Now actually read all the data
@@ -289,130 +302,13 @@ class TdmsFile(object):
 
         self.data_read = True
 
-    def _read_channel_data_chunks(self, channel):
-        reader = self._get_reader()
-        for chunk in reader.read_raw_data_for_channel(channel.path):
-            yield chunk
-
-    def _read_channel_data_chunk_for_index(self, channel, index):
-        return self._get_reader().read_channel_chunk_for_index(channel.path, index)
-
-    def _read_channel_data(self, channel, offset=0, length=None):
-        if offset < 0:
-            raise ValueError("offset must be non-negative")
-        if length is not None and length < 0:
-            raise ValueError("length must be non-negative")
-        reader = self._get_reader()
-
-        with Timer(log, "Allocate space for channel"):
-            # Allocate space for data
-            if length is None:
-                num_values = len(channel) - offset
-            else:
-                num_values = min(length, len(channel) - offset)
-            num_values = max(0, num_values)
-            channel_data = get_data_receiver(channel, num_values, self._memmap_dir)
-
-        with Timer(log, "Read data for channel"):
-            # Now actually read all the data
-            for chunk in reader.read_raw_data_for_channel(channel.path, offset, length):
-                if chunk.data is not None:
-                    channel_data.append_data(chunk.data)
-                if chunk.scaler_data is not None:
-                    for scaler_id, scaler_data in chunk.scaler_data.items():
-                        channel_data.append_scaler_data(scaler_id, scaler_data)
-
-        return channel_data
-
-    def object(self, *path):
-        """(Deprecated) Get a TDMS object from the file
-
-        :param path: The object group and channel names. Providing no channel
-            returns a group object, and providing no channel or group
-            will return the root object.
-        :rtype: One of :class:`TdmsGroup`, :class:`TdmsChannel`, :class:`RootObject`
-
-        For example, to get the root object::
-
-            object()
-
-        To get a group::
-
-            object("group_name")
-
-        To get a channel::
-
-            object("group_name", "channel_name")
-        """
-
-        _deprecated("TdmsFile.object",
-                    "Use TdmsFile.properties to access properties of the root object, " +
-                    "TdmsFile[group_name] to access a group object and " +
-                    "TdmsFile[group_name][channel_name] to access a channel object.")
-
-        def get_name(component):
-            try:
-                return component.name
-            except AttributeError:
-                return component
-
-        path = [get_name(c) for c in path]
-        object_path = ObjectPath(*path)
-        try:
-            return self.objects[str(object_path)]
-        except KeyError:
-            raise KeyError("Invalid object path: %s" % object_path)
-
-    def group_channels(self, group):
-        """(Deprecated) Returns a list of channel objects for the given group
-
-        :param group: Group or name of the group to get channels for.
-        :rtype: List of :class:`TdmsObject` objects.
-        """
-
-        _deprecated("TdmsFile.group_channels", "Use TdmsFile[group_name].channels().")
-
-        if isinstance(group, TdmsGroup):
-            return group.channels()
-
-        return self._groups[group].channels()
-
-    def channel_data(self, group, channel):
-        """(Deprecated)  Get the data for a channel
-
-        :param group: The name of the group the channel is in.
-        :param channel: The name of the channel to get data for.
-        :returns: The channel data.
-        :rtype: NumPy array.
-        """
-
-        _deprecated("TdmsFile.channel_data", "Use TdmsFile[group_name][channel_name].data.")
-
-        if self._reader is None:
-            # Data should have already been loaded
-            return self[group][channel].data
-        else:
-            # Data must be lazily loaded
-            return self[group][channel].read_data()
-
-    @_property_builtin
-    def objects(self):
-        """ (Deprecated) A dictionary of objects in the TDMS file, where the keys are the object paths.
-        """
-
-        _deprecated("TdmsFile.objects", "Use TdmsFile.groups() to access all groups in the file, " +
-                    "and group.channels() to access all channels in a group.")
-
-        objects = OrderedDict()
-        root_path = ObjectPath()
-        objects[str(root_path)] = RootObject(self._properties)
-
-        for group in self.groups():
-            objects[group.path] = group
-            for channel in group.channels():
-                objects[channel.path] = channel
-
-        return objects
+    def _convert_properties(self, properties):
+        def convert_prop(val):
+            if isinstance(val, TdmsTimestamp) and not self._raw_timestamps:
+                # Convert timestamps to numpy datetime64 if raw timestamps are not requested
+                return val.as_datetime64()
+            return val
+        return OrderedDict((k, convert_prop(v)) for (k, v) in properties.items())
 
 
 class TdmsGroup(object):
@@ -442,13 +338,13 @@ class TdmsGroup(object):
     def __repr__(self):
         return "<TdmsGroup with path %s>" % self.path
 
-    @_property_builtin
+    @property
     def path(self):
         """ Path to the TDMS object for this group
         """
         return str(self._path)
 
-    @_property_builtin
+    @property
     def name(self):
         """ The name of this group
         """
@@ -497,44 +393,6 @@ class TdmsGroup(object):
                 "There is no channel named '%s' in group '%s' of the TDMS file" %
                 (channel_name, self.name))
 
-    def property(self, property_name):
-        """(Deprecated) Returns the value of a TDMS property
-
-        :param property_name: The name of the property to get.
-        :returns: The value of the requested property.
-        :raises: KeyError if the property isn't found.
-        """
-
-        _deprecated("TdmsGroup.property", "Use TdmsGroup.properties[property_name].")
-
-        try:
-            return self.properties[property_name]
-        except KeyError:
-            raise KeyError(
-                "Object does not have property '%s'" % property_name)
-
-    @_property_builtin
-    def group(self):
-        """ (Deprecated) Returns the name of the group for this object,
-            or None if it is the root object.
-        """
-        _deprecated("TdmsGroup.group", "Use TdmsGroup.name.")
-        return self._path.group
-
-    @_property_builtin
-    def channel(self):
-        """ (Deprecated) Returns the name of the channel for this object,
-            or None if it is a group or the root object.
-        """
-        _deprecated("TdmsGroup.channel", "This always returns None.")
-        return None
-
-    @_property_builtin
-    def has_data(self):
-        """(Deprecated)"""
-        _deprecated("TdmsGroup.has_data", "This always returns False.")
-        return False
-
 
 class TdmsChannel(object):
     """ Represents a data channel in a TDMS file.
@@ -562,14 +420,19 @@ class TdmsChannel(object):
     """
 
     def __init__(
-            self, tdms_file, path, properties, data_type=None,
-            scaler_data_types=None, number_values=0):
-        self._tdms_file = tdms_file
+            self, path, data_type, scaler_data_types, number_values,
+            properties, group_properties, file_properties,
+            tdms_reader, raw_timestamps, memmap_dir):
         self._path = path
         self.properties = properties
         self._length = number_values
         self.data_type = data_type
         self.scaler_data_types = scaler_data_types
+        self._group_properties = group_properties
+        self._file_properties = file_properties
+        self._reader = tdms_reader
+        self._raw_timestamps = raw_timestamps
+        self._memmap_dir = memmap_dir
 
         self._raw_data = None
         self._cached_chunk = None
@@ -603,18 +466,24 @@ class TdmsChannel(object):
         else:
             raise TypeError("Invalid index type '%s', expected int, slice or Ellipsis" % type(index).__name__)
 
-    @_property_builtin
+    @property
     def path(self):
         """ Path to the TDMS object for this channel
         """
         return str(self._path)
 
-    @_property_builtin
+    @property
     def name(self):
         """ The name of this channel
         """
         return self._path.channel
 
+    @property
+    def group_name(self):
+        """ The name of the group that contains this channel
+        """
+        return self._path.group
+
     @cached_property
     def dtype(self):
         """ NumPy data type of the channel data
@@ -653,7 +522,7 @@ class TdmsChannel(object):
             return np.empty((0, ), dtype=self.dtype)
         return self._scale_data(self._raw_data)
 
-    @_property_builtin
+    @property
     def raw_data(self):
         """ If the TdmsFile was created by reading all data, this property
         provides direct access to the numpy array of raw, unscaled data.
@@ -674,7 +543,7 @@ class TdmsChannel(object):
                     "for a scale_id")
         return self._raw_data.data
 
-    @_property_builtin
+    @property
     def raw_scaler_data(self):
         """ If the TdmsFile was created by reading all data, this property
         provides direct access to the numpy array of raw DAQmx scaler data
@@ -692,8 +561,8 @@ class TdmsChannel(object):
         :rtype: Generator that yields :class:`ChannelDataChunk` objects
         """
         channel_offset = 0
-        for raw_data_chunk in self._tdms_file._read_channel_data_chunks(self):
-            yield ChannelDataChunk(self._tdms_file, self, raw_data_chunk, channel_offset)
+        for raw_data_chunk in self._read_channel_data_chunks():
+            yield ChannelDataChunk(self, raw_data_chunk, channel_offset)
             channel_offset += len(raw_data_chunk)
 
     def read_data(self, offset=0, length=None, scaled=True):
@@ -709,7 +578,11 @@ class TdmsChannel(object):
             Set this parameter to False to return raw unscaled data.
             For DAQmx data a dictionary of scaler id to raw scaler data will be returned.
         """
-        raw_data = self._tdms_file._read_channel_data(self, offset, length)
+        if self._raw_data is None:
+            raw_data = self._read_channel_data(offset, length)
+        else:
+            raw_data = slice_raw_data(self._raw_data, offset, length)
+
         if raw_data is None:
             dtype = self.dtype if scaled else self._raw_data_dtype()
             return np.empty((0,), dtype=dtype)
@@ -764,6 +637,8 @@ class TdmsChannel(object):
         except KeyError:
             raise KeyError(
                 "Object does not have start time property available.")
+        if isinstance(start_time, TdmsTimestamp):
+            start_time = start_time.as_datetime64(accuracy)
 
         try:
             unit_correction = {
@@ -778,7 +653,7 @@ class TdmsChannel(object):
         # Because numpy only knows ints as its date datatype,
         # convert to accuracy.
         time_type = "timedelta64[{0}]".format(accuracy)
-        return (np.datetime64(start_time) +
+        return (start_time +
                 (relative_time * unit_correction).astype(time_type))
 
     def as_dataframe(self, time_index=False, absolute_time=False, scaled_data=True):
@@ -846,8 +721,11 @@ class TdmsChannel(object):
             return read_data[::step]
 
     def _read_at_index(self, index):
+        orig_index = index
+        if index < 0:
+            index = self._length + index
         if index < 0 or index >= self._length:
-            raise IndexError("Index {0} is outside of the channel bounds [0, {1}]".format(index, self._length - 1))
+            raise IndexError("Index {0} is outside of the channel bounds [0, {1}]".format(orig_index, self._length - 1))
 
         if self._cached_chunk is not None:
             # Check if we've already read and cached the chunk containing this index
@@ -855,7 +733,7 @@ class TdmsChannel(object):
             if bounds[0] <= index < bounds[1]:
                 return self._cached_chunk[index - bounds[0]]
 
-        chunk, chunk_offset = self._tdms_file._read_channel_data_chunk_for_index(self, index)
+        chunk, chunk_offset = self._read_channel_data_chunk_for_index(index)
         scaled_chunk = self._scale_data(chunk)
         self._cached_chunk = scaled_chunk
         self._cached_chunk_bounds = (chunk_offset, chunk_offset + len(scaled_chunk))
@@ -873,57 +751,48 @@ class TdmsChannel(object):
 
     @cached_property
     def _scaling(self):
-        group_properties = self._tdms_file[self._path.group].properties
-        file_properties = self._tdms_file.properties
         return scaling.get_scaling(
-            self.properties, group_properties, file_properties)
+            self.properties, self._group_properties, self._file_properties)
+
+    def _read_channel_data_chunks(self):
+        for chunk in self._reader.read_raw_data_for_channel(self.path):
+            _convert_channel_data_chunk(chunk, self._raw_timestamps)
+            yield chunk
+
+    def _read_channel_data_chunk_for_index(self, index):
+        (chunk, offset) = self._reader.read_channel_chunk_for_index(self.path, index)
+        _convert_channel_data_chunk(chunk, self._raw_timestamps)
+        return chunk, offset
+
+    def _read_channel_data(self, offset=0, length=None):
+        if offset < 0:
+            raise ValueError("offset must be non-negative")
+        if length is not None and length < 0:
+            raise ValueError("length must be non-negative")
+
+        with Timer(log, "Allocate space for channel"):
+            # Allocate space for data
+            if length is None:
+                num_values = len(self) - offset
+            else:
+                num_values = min(length, len(self) - offset)
+            num_values = max(0, num_values)
+            channel_data = get_data_receiver(self, num_values, self._raw_timestamps, self._memmap_dir)
+
+        with Timer(log, "Read data for channel"):
+            # Now actually read all the data
+            for chunk in self._reader.read_raw_data_for_channel(self.path, offset, length):
+                if chunk.data is not None:
+                    channel_data.append_data(chunk.data)
+                if chunk.scaler_data is not None:
+                    for scaler_id, scaler_data in chunk.scaler_data.items():
+                        channel_data.append_scaler_data(scaler_id, scaler_data)
+
+        return channel_data
 
     def _set_raw_data(self, data):
         self._raw_data = data
 
-    def property(self, property_name):
-        """(Deprecated) Returns the value of a TDMS property
-
-        :param property_name: The name of the property to get.
-        :returns: The value of the requested property.
-        :raises: KeyError if the property isn't found.
-        """
-        _deprecated("TdmsChannel.property", "Use TdmsChannel.properties[property_name]")
-
-        try:
-            return self.properties[property_name]
-        except KeyError:
-            raise KeyError(
-                "Object does not have property '%s'" % property_name)
-
-    @_property_builtin
-    def group(self):
-        """ (Deprecated) Returns the name of the group for this object,
-            or None if it is the root object.
-        """
-        _deprecated("TdmsChannel.group")
-        return self._path.group
-
-    @_property_builtin
-    def channel(self):
-        """ (Deprecated) Returns the name of the channel for this object,
-            or None if it is a group or the root object.
-        """
-        _deprecated("TdmsChannel.channel", "Use TdmsChannel.name")
-        return self._path.channel
-
-    @_property_builtin
-    def has_data(self):
-        """(Deprecated)"""
-        _deprecated("TdmsChannel.has_data", "This always returns True")
-        return True
-
-    @_property_builtin
-    def number_values(self):
-        """(Deprecated)"""
-        _deprecated("TdmsChannel.number_values", "Use len(channel)")
-        return self._length
-
 
 class DataChunk(object):
     """ A chunk of data in a TDMS file
@@ -967,7 +836,6 @@ class GroupDataChunk(object):
         self.name = group.name
         self._channels = OrderedDict(
             (channel.name, ChannelDataChunk(
-                tdms_file,
                 channel,
                 raw_data_chunk.channel_data.get(channel.path, RawChannelDataChunk.empty()),
                 channel_offsets[channel.path]))
@@ -997,9 +865,8 @@ class ChannelDataChunk(object):
     :ivar ~.name: Name of the channel
     :ivar ~.offset: Starting index of this chunk of data in the entire channel
     """
-    def __init__(self, tdms_file, channel, raw_data_chunk, offset):
+    def __init__(self, channel, raw_data_chunk, offset):
         self._path = channel._path
-        self._tdms_file = tdms_file
         self._channel = channel
         self.name = channel.name
         self.offset = offset
@@ -1033,36 +900,11 @@ class ChannelDataChunk(object):
             return self._raw_data.data
 
 
-class RootObject(object):
-    def __init__(self, properties):
-        self.properties = properties
-
-    def property(self, property_name):
-        _deprecated("RootObject", "Use TdmsFile.properties to access properties from the root object")
-        try:
-            return self.properties[property_name]
-        except KeyError:
-            raise KeyError(
-                "Object does not have property '%s'" % property_name)
-
-    @_property_builtin
-    def group(self):
-        _deprecated("RootObject", "Use TdmsFile.properties to access properties from the root object")
-        return None
-
-    @_property_builtin
-    def channel(self):
-        _deprecated("RootObject", "Use TdmsFile.properties to access properties from the root object")
-        return None
-
-    @_property_builtin
-    def has_data(self):
-        _deprecated("RootObject", "Use TdmsFile.properties to access properties from the root object")
-        return False
+def _convert_data_chunk(chunk, raw_timestamps):
+    for channel_chunk in chunk.channel_data.values():
+        _convert_channel_data_chunk(channel_chunk, raw_timestamps)
 
 
-def _deprecated(name, detail=None):
-    message = "'{0}' is deprecated and will be removed in a future release.".format(name)
-    if detail is not None:
-        message += " {0}".format(detail)
-    warnings.warn(message)
+def _convert_channel_data_chunk(channel_chunk, raw_timestamps):
+    if not raw_timestamps and isinstance(channel_chunk.data, TimestampArray):
+        channel_chunk.data = channel_chunk.data.as_datetime64()
