@@ -10,7 +10,9 @@ from nptdms.test.util import (
     segment_objects_metadata,
     channel_metadata,
     channel_metadata_with_no_data,
-    channel_metadata_with_repeated_structure)
+    channel_metadata_with_repeated_structure,
+    root_metadata,
+    group_metadata)
 
 
 TDS_TYPE_INT8 = 1
@@ -46,12 +48,12 @@ def single_segment_with_one_channel():
     test_file.add_segment(
         ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
         segment_objects_metadata(
-            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
+            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 4),
         ),
-        "01 00 00 00" "02 00 00 00"
+        "01 00 00 00" "02 00 00 00" "03 00 00 00" "04 00 00 00"
     )
     expected_data = {
-        ('group', 'channel1'): np.array([1, 2], dtype=np.int32),
+        ('group', 'channel1'): np.array([1, 2, 3, 4], dtype=np.int32),
     }
     return test_file, expected_data
 
@@ -538,24 +540,30 @@ def incomplete_last_segment():
     test_file.add_segment(
         ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
         segment_objects_metadata(
+            root_metadata(),
+            group_metadata(),
             channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
             channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
+            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 2),
         ),
         "01 00 00 00" "02 00 00 00"
         "03 00 00 00" "04 00 00 00"
         "05 00 00 00" "06 00 00 00"
         "07 00 00 00" "08 00 00 00"
+        "09 00 00 00" "0A 00 00 00"
+        "0B 00 00 00" "0C 00 00 00"
     )
     test_file.add_segment(
         ("kTocRawData", ),
         "",
-        "09 00 00 00" "0A 00 00 00"
-        "0B 00 00 00" "0C 00 00 00",
+        "01 00 00 00" "02 00 00 00"
+        "03 00 00 00",
         incomplete=True
     )
     expected_data = {
-        ('group', 'channel1'): np.array([1, 2, 5, 6, 9, 10], dtype=np.int32),
-        ('group', 'channel2'): np.array([3, 4, 7, 8, 11, 12], dtype=np.int32),
+        ('group', 'channel1'): np.array([1, 2, 7, 8, 1, 2], dtype=np.int32),
+        ('group', 'channel2'): np.array([3, 4, 9, 10, 3], dtype=np.int32),
+        ('group', 'channel3'): np.array([5, 6, 11, 12], dtype=np.int32),
     }
     return test_file, expected_data
 
@@ -568,6 +576,8 @@ def incomplete_last_row_of_interleaved_data():
     test_file.add_segment(
         ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
         segment_objects_metadata(
+            root_metadata(),
+            group_metadata(),
             channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
             channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
         ),
@@ -740,43 +750,15 @@ def timestamp_data():
         np.datetime64('2012-08-23T12:02:03.9999', 'us'),
     ]
 
-    metadata = (
-        # Number of objects
-        "02 00 00 00"
-        # Length of the object path
-        "17 00 00 00")
-    metadata += string_hexlify("/'Group'/'TimeChannel1'")
-    metadata += (
-        # Length of index information
-        "14 00 00 00"
-        # Raw data data type
-        "44 00 00 00"
-        # Dimension
-        "01 00 00 00"
-        # Number of raw data values
-        "02 00 00 00"
-        "00 00 00 00"
-        # Number of properties (0)
-        "00 00 00 00")
-    metadata += (
-        "17 00 00 00")
-    metadata += string_hexlify("/'Group'/'TimeChannel2'")
-    metadata += (
-        # Length of index information
-        "14 00 00 00"
-        # Raw data data type
-        "44 00 00 00"
-        # Dimension
-        "01 00 00 00"
-        # Number of raw data values
-        "02 00 00 00"
-        "00 00 00 00"
-        # Number of properties (0)
-        "00 00 00 00")
-
     test_file = GeneratedFile()
-    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
-    test_file.add_segment(toc, metadata, timestamp_data_chunk(times))
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'Group'/'TimeChannel1'", 0x44, 2),
+            channel_metadata("/'Group'/'TimeChannel2'", 0x44, 2),
+        ),
+        timestamp_data_chunk(times)
+    )
 
     expected_data = {
         ('Group', 'TimeChannel1'): np.array([times[0], times[1]]),
@@ -974,25 +956,41 @@ def channel_without_data_or_data_type():
     return test_file, expected_data
 
 
+@scenario
+def extra_padding_after_metadata():
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
+        ) + "00 00 00 00 00 00 00 00",
+        "01 00 00 00" "02 00 00 00"
+    )
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
+        ) + "00 00 00 00 00 00 00 00",
+        "03 00 00 00" "04 00 00 00"
+    )
+    expected_data = {
+        ('group', 'channel1'): np.array([1, 2, 3, 4], dtype=np.int32),
+    }
+    return test_file, expected_data
+
+
 def timestamp_data_chunk(times):
+    return "".join(timestamp_hexlify(time) for time in times)
+
+
+def timestamp_hexlify(time):
     epoch = np.datetime64('1904-01-01T00:00:00')
 
-    def total_seconds(td):
-        return int(td / np.timedelta64(1, 's'))
+    relative_time = time - epoch
+    total_seconds = int(relative_time / np.timedelta64(1, 's'))
 
-    def microseconds(dt):
-        diff = dt - epoch
-        secs = total_seconds(diff)
-        remainder = diff - np.timedelta64(secs, 's')
-        return int(remainder / np.timedelta64(1, 'us'))
+    remainder = relative_time - np.timedelta64(total_seconds, 's')
+    microseconds = int(remainder / np.timedelta64(1, 'us'))
+    fraction = int(float(microseconds) * 2 ** 58 / 5 ** 6)
 
-    seconds = [total_seconds(t - epoch) for t in times]
-    fractions = [
-        int(float(microseconds(t)) * 2 ** 58 / 5 ** 6)
-        for t in times]
-
-    data = ""
-    for f, s in zip(fractions, seconds):
-        data += hexlify_value("<Q", f)
-        data += hexlify_value("<q", s)
-    return data
+    return hexlify_value("<Q", fraction) + hexlify_value("<q", total_seconds)
