@@ -6,15 +6,6 @@ import pytest
 from nptdms import types
 from nptdms.scaling import get_scaling
 
-try:
-    import thermocouples_reference
-except ImportError:
-    thermocouples_reference = None
-try:
-    import scipy
-except ImportError:
-    scipy = None
-
 
 def test_unsupported_scaling_type():
     """Raw data is returned unscaled when the scaling type is unsupported.
@@ -47,7 +38,7 @@ def test_linear_scaling():
 
     assert scaling.get_dtype(types.Int32, None) == np.dtype('float64')
     assert scaled_data.dtype == np.dtype('float64')
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_polynomial_scaling():
@@ -69,7 +60,7 @@ def test_polynomial_scaling():
 
     assert scaling.get_dtype(types.Int32, None) == np.dtype('float64')
     assert scaled_data.dtype == np.dtype('float64')
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_polynomial_scaling_with_no_coefficients():
@@ -86,7 +77,7 @@ def test_polynomial_scaling_with_no_coefficients():
     scaling = get_scaling(properties, {}, {})
     scaled_data = scaling.scale(data)
 
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_polynomial_scaling_with_3_coefficients():
@@ -106,7 +97,7 @@ def test_polynomial_scaling_with_3_coefficients():
     scaling = get_scaling(properties, {}, {})
     scaled_data = scaling.scale(data)
 
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 @pytest.mark.parametrize(
@@ -141,7 +132,37 @@ def test_rtd_scaling(resistance_configuration, lead_resistance, expected_data):
     scaled_data = scaling.scale(data)
 
     assert scaling.get_dtype(types.DoubleFloat, None) == np.dtype('float64')
-    np.testing.assert_almost_equal(expected_data, scaled_data, decimal=3)
+    np.testing.assert_almost_equal(scaled_data, expected_data, decimal=3)
+
+
+def test_rtd_scaling_with_negative_temperature():
+    """ Test RTD scaling with negative temperature values, which requires
+        solving the full quartic Callendar-Van Dusen equation
+    """
+    data = StubTdmsData(np.array([
+        0.08, 0.09, 0.095, 0.1, 0.12, 0.14, 0.16, 0.18, 0.2, 0.22, 0.24]))
+    expected_data = np.array([
+        -50.77114, -25.48835, -12.76894, -0., 51.56605, 103.94273,
+        157.1695, 211.28915, 266.34819, 322.3973, 379.49189])
+
+    properties = {
+        "NI_Number_Of_Scales": 1,
+        "NI_Scale[0]_Scale_Type": "RTD",
+        "NI_Scale[0]_RTD_Current_Excitation": 0.001,
+        "NI_Scale[0]_RTD_R0_Nominal_Resistance": 100.0,
+        "NI_Scale[0]_RTD_A": 0.0039083,
+        "NI_Scale[0]_RTD_B": -5.775e-07,
+        "NI_Scale[0]_RTD_C": -4.183e-12,
+        "NI_Scale[0]_RTD_Lead_Wire_Resistance": 0,
+        "NI_Scale[0]_RTD_Resistance_Configuration": 2,
+        "NI_Scale[0]_RTD_Input_Source": 0xFFFFFFFF,
+    }
+    scaling = get_scaling(properties, {}, {})
+    scaled_data = scaling.scale(data)
+
+    assert scaling.get_dtype(types.DoubleFloat, None) == np.dtype('float64')
+    assert scaled_data.dtype == np.dtype('float64')
+    np.testing.assert_almost_equal(scaled_data, expected_data, decimal=5)
 
 
 def test_table_scaling():
@@ -170,7 +191,7 @@ def test_table_scaling():
     scaled_data = scaling.scale(data)
 
     assert scaling.get_dtype(types.DoubleFloat, None) == np.dtype('float64')
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_add_scaling():
@@ -193,7 +214,7 @@ def test_add_scaling():
 
     assert scaling.get_dtype(None, {0: types.Int32, 1: types.Uint32}) == np.dtype('int64')
     assert scaled_data.dtype == np.dtype('int64')
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_add_scaling_with_default_inputs():
@@ -238,16 +259,14 @@ def test_subtract_scaling():
 
     assert scaling.get_dtype(None, {0: types.Int32, 1: types.Uint32}) == np.dtype('int64')
     assert scaled_data.dtype == np.dtype('int64')
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
-@pytest.mark.skipif(thermocouples_reference is None, reason="thermocouples_reference is not installed")
-@pytest.mark.skipif(scipy is None, reason="scipy is not installed")
 def test_thermocouple_scaling_voltage_to_temperature():
     """Test thermocouple scaling from a voltage in uV to temperature"""
 
     data = StubTdmsData(np.array([0.0, 10.0, 100.0, 1000.0]))
-    expected_scaled_data = np.array([0.0, 0.2534448,  2.5309141, 24.9940185])
+    expected_scaled_data = np.array([0.000000, 0.250843, 2.508899, 24.983648])
 
     properties = {
         "NI_Number_Of_Scales": 1,
@@ -260,10 +279,9 @@ def test_thermocouple_scaling_voltage_to_temperature():
     scaled_data = scaling.scale(data)
 
     np.testing.assert_almost_equal(
-        expected_scaled_data, scaled_data, decimal=3)
+        scaled_data, expected_scaled_data, decimal=3)
 
 
-@pytest.mark.skipif(thermocouples_reference is None, reason="thermocouples_reference is not installed")
 def test_thermocouple_scaling_temperature_to_voltage():
     """Test thermocouple scaling from a temperature to voltage in uV"""
 
@@ -282,7 +300,48 @@ def test_thermocouple_scaling_temperature_to_voltage():
     scaled_data = scaling.scale(data)
 
     np.testing.assert_almost_equal(
-        expected_scaled_data, scaled_data, decimal=3)
+        scaled_data, expected_scaled_data, decimal=3)
+
+
+def test_thermocouple_scaling_voltage_to_temperature_benchmark(benchmark):
+    """Test thermocouple scaling from a voltage in uV to temperature"""
+
+    data = StubTdmsData(np.tile(np.array([0.0, 10.0, 100.0, 1000.0]), 100))
+    expected_scaled_data = np.tile(np.array([0.000000, 0.250843, 2.508899, 24.983648]), 100)
+
+    properties = {
+        "NI_Number_Of_Scales": 1,
+        "NI_Scale[0]_Scale_Type": "Thermocouple",
+        "NI_Scale[0]_Thermocouple_Thermocouple_Type": 10073,
+        "NI_Scale[0]_Thermocouple_Scaling_Direction": 0,
+        "NI_Scale[0]_Thermocouple_Input_Source": 0xFFFFFFFF,
+    }
+    scaling = get_scaling(properties, {}, {})
+    scaled_data = benchmark(scaling.scale, data)
+
+    np.testing.assert_almost_equal(
+        scaled_data, expected_scaled_data, decimal=3)
+
+
+def test_thermocouple_scaling_temperature_to_voltage_benchmark(benchmark):
+    """Test thermocouple scaling from a temperature to voltage in uV"""
+
+    data = StubTdmsData(np.tile(np.array([0.0, 10.0, 50.0, 100.0]), 100))
+    expected_scaled_data = np.tile(np.array([
+        0.0, 396.8619078, 2023.0778862, 4096.2302187]), 100)
+
+    properties = {
+        "NI_Number_Of_Scales": 1,
+        "NI_Scale[0]_Scale_Type": "Thermocouple",
+        "NI_Scale[0]_Thermocouple_Thermocouple_Type": 10073,
+        "NI_Scale[0]_Thermocouple_Scaling_Direction": 1,
+        "NI_Scale[0]_Thermocouple_Input_Source": 0xFFFFFFFF,
+    }
+    scaling = get_scaling(properties, {}, {})
+    scaled_data = benchmark(scaling.scale, data)
+
+    np.testing.assert_almost_equal(
+        scaled_data, expected_scaled_data, decimal=3)
 
 
 @pytest.mark.parametrize(
@@ -317,7 +376,7 @@ def test_thermistor_scaling_with_voltage_excitation(
     scaling = get_scaling(properties, {}, {})
     scaled_data = scaling.scale(data)
 
-    np.testing.assert_almost_equal(expected_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_data)
 
 
 @pytest.mark.parametrize(
@@ -352,7 +411,7 @@ def test_thermistor_scaling_with_current_excitation(
     scaling = get_scaling(properties, {}, {})
     scaled_data = scaling.scale(data)
 
-    np.testing.assert_almost_equal(expected_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_data)
 
 
 def test_thermistor_scaling_with_invalid_excitation_type():
@@ -378,6 +437,80 @@ def test_thermistor_scaling_with_invalid_excitation_type():
     assert "Invalid excitation type: 12345" in str(exc_info.value)
 
 
+def test_strain_scaling_full_bridge_type_i():
+    _strain_scaling_test(
+        [0.00068827, 0.00068036, 0.000688, 0.00068545, 0.00069104,
+         0.00068033, 0.00068023, 0.00068316, 0.00067672, 0.000679],
+        [-0.0001311, -0.00012959, -0.00013105, -0.00013056, -0.00013163,
+         -0.00012959, -0.00012957, -0.00013013, -0.0001289, -0.00012933])
+
+
+def test_strain_scaling_full_bridge_type_i_with_adjustment():
+    _strain_scaling_test(
+        [0.00068827, 0.00068036, 0.000688, 0.00068545, 0.00069104,
+         0.00068033, 0.00068023, 0.00068316, 0.00067672, 0.000679],
+        [-0.00014722, -0.00014553, -0.00014717, -0.00014662, -0.00014782,
+         -0.00014553, -0.0001455, -0.00014613, -0.00014475, -0.00014524],
+        calibration_adjustment=1.123)
+
+
+def test_strain_scaling_full_bridge_type_i_with_initial_voltage():
+    _strain_scaling_test(
+        [0.00068827, 0.00068036, 0.000688, 0.00068545, 0.00069104,
+         0.00068033, 0.00068023, 0.00068316, 0.00067672, 0.000679],
+        [0.00012604, 0.00012755, 0.0001261, 0.00012658, 0.00012552,
+         0.00012756, 0.00012758, 0.00012702, 0.00012824, 0.00012781],
+        initial_bridge_voltage=0.00135)
+
+
+def _strain_scaling_test(input_data, expected_data, initial_bridge_voltage=0.0, calibration_adjustment=1.0):
+    data = StubTdmsData(np.array(input_data))
+    expected_data = np.array(expected_data)
+
+    properties = {
+        "NI_Number_Of_Scales": 1,
+        "NI_Scale[0]_Scale_Type": "Strain",
+        "NI_Scale[0]_Strain_Configuration": 10183,
+        "NI_Scale[0]_Strain_Poisson_Ratio": 0.3,
+        "NI_Scale[0]_Strain_Gage_Resistance": 350.0,
+        "NI_Scale[0]_Strain_Lead_Wire_Resistance": 0.0,
+        "NI_Scale[0]_Strain_Initial_Bridge_Voltage": initial_bridge_voltage,
+        "NI_Scale[0]_Strain_Gage_Factor": 2.1,
+        "NI_Scale[0]_Strain_Bridge_Shunt_Calibration_Gain_Adjustment": calibration_adjustment,
+        "NI_Scale[0]_Strain_Voltage_Excitation": 2.5,
+        "NI_Scale[0]_Strain_Input_Source": 0xFFFFFFFF,
+    }
+
+    scaling = get_scaling(properties, {}, {})
+    scaled_data = scaling.scale(data)
+
+    np.testing.assert_almost_equal(scaled_data, expected_data)
+
+
+def test_unsupported_strain_configuration():
+    data = StubTdmsData(np.array([0.0, 0.0, 0.0]))
+
+    properties = {
+        "NI_Number_Of_Scales": 1,
+        "NI_Scale[0]_Scale_Type": "Strain",
+        "NI_Scale[0]_Strain_Configuration": 10184,
+        "NI_Scale[0]_Strain_Poisson_Ratio": 0.3,
+        "NI_Scale[0]_Strain_Gage_Resistance": 350.0,
+        "NI_Scale[0]_Strain_Lead_Wire_Resistance": 0.0,
+        "NI_Scale[0]_Strain_Initial_Bridge_Voltage": 0.0,
+        "NI_Scale[0]_Strain_Gage_Factor": 2.1,
+        "NI_Scale[0]_Strain_Bridge_Shunt_Calibration_Gain_Adjustment": 1.0,
+        "NI_Scale[0]_Strain_Voltage_Excitation": 2.5,
+        "NI_Scale[0]_Strain_Input_Source": 0xFFFFFFFF,
+    }
+
+    scaling = get_scaling(properties, {}, {})
+
+    with pytest.raises(Exception) as exc_info:
+        _ = scaling.scale(data)
+    assert str(exc_info.value) == "Strain gauge configuration 10184 is not supported"
+
+
 def test_multiple_scalings_applied_in_order():
     """Test all scalings applied from multiple scalings
     """
@@ -404,7 +537,7 @@ def test_multiple_scalings_applied_in_order():
     scaling = get_scaling(properties, {}, {})
     scaled_data = scaling.scale(data)
 
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_multiple_scalings_but_all_with_raw_data_input():
@@ -433,7 +566,7 @@ def test_multiple_scalings_but_all_with_raw_data_input():
     scaling = get_scaling(properties, {}, {})
     scaled_data = scaling.scale(data)
 
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_scaling_from_group():
@@ -451,7 +584,7 @@ def test_scaling_from_group():
     scaling = get_scaling({}, group_properties, {})
     scaled_data = scaling.scale(data)
 
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_scaling_from_root():
@@ -469,7 +602,7 @@ def test_scaling_from_root():
     scaling = get_scaling({}, {}, root_properties)
     scaled_data = scaling.scale(data)
 
-    np.testing.assert_almost_equal(expected_scaled_data, scaled_data)
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
 
 
 def test_scaling_status_scaled():
@@ -486,6 +619,24 @@ def test_scaling_status_scaled():
     assert scaling is None
 
 
+def test_advanced_api_scaling():
+    """Test AdvancedAPI scaling (no-op)"""
+
+    data = StubTdmsData(np.array([1, 2, 3], dtype=np.dtype('int32')))
+    expected_scaled_data = np.array([1, 2, 3])
+
+    properties = {
+        "NI_Number_Of_Scales": 1,
+        "NI_Scale[0]_Scale_Type": "AdvancedAPI",
+    }
+    scaling = get_scaling(properties, {}, {})
+    scaled_data = scaling.scale(data)
+
+    assert scaling.get_dtype(types.Int32, None) == np.dtype('int32')
+    assert scaled_data.dtype == np.dtype('int32')
+    np.testing.assert_almost_equal(scaled_data, expected_scaled_data)
+
+
 class StubTdmsData(object):
     def __init__(self, data):
         self.data = data
