@@ -4,8 +4,10 @@ from collections import defaultdict
 import logging
 import os
 import sys
+from shutil import copyfile
 import tempfile
-from hypothesis import (assume, given, example, settings, strategies)
+import weakref
+from hypothesis import (assume, given, example, settings, strategies, HealthCheck)
 import numpy as np
 import pytest
 from nptdms import TdmsFile
@@ -23,6 +25,12 @@ from nptdms.test.util import (
 from nptdms.test import scenarios
 
 
+# When running tests on GitHub actions, the first iteration can be quite
+# slow and cause failures, so disable deadlines:
+settings.register_profile('ci', deadline=None)
+settings.load_profile('ci')
+
+
 @pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
 def test_read_channel_data(test_file, expected_data):
     """Test reading data"""
@@ -50,6 +58,18 @@ def test_lazily_read_channel_data(test_file, expected_data):
                 compare_arrays(actual_data, expected_data)
 
 
+def test_read_raw_channel_data():
+    """Test reading raw channel data"""
+
+    test_file, expected_data = scenarios.single_segment_with_one_channel().values
+    with test_file.get_tempfile() as temp_file:
+        tdms_file = TdmsFile.read(temp_file.file)
+        for ((group, channel), expected_data) in expected_data.items():
+            actual_data = tdms_file[group][channel].read_data(scaled=False)
+            assert actual_data.dtype == expected_data.dtype
+            compare_arrays(actual_data, expected_data)
+
+
 def test_lazily_read_raw_channel_data():
     """Test reading raw channel data lazily"""
 
@@ -62,6 +82,30 @@ def test_lazily_read_raw_channel_data():
                 compare_arrays(actual_data, expected_data)
 
 
+def test_read_raw_channel_data_slice():
+    """Test reading a slice of raw channel data"""
+
+    test_file, expected_data = scenarios.single_segment_with_one_channel().values
+    with test_file.get_tempfile() as temp_file:
+        tdms_file = TdmsFile.read(temp_file.file)
+        for ((group, channel), expected_data) in expected_data.items():
+            actual_data = tdms_file[group][channel].read_data(offset=1, length=2, scaled=False)
+            assert actual_data.dtype == expected_data.dtype
+            compare_arrays(actual_data, expected_data[1:3])
+
+
+def test_lazily_read_raw_channel_data_slice():
+    """Test reading raw channel data lazily"""
+
+    test_file, expected_data = scenarios.single_segment_with_one_channel().values
+    with test_file.get_tempfile() as temp_file:
+        with TdmsFile.open(temp_file.file) as tdms_file:
+            for ((group, channel), expected_data) in expected_data.items():
+                actual_data = tdms_file[group][channel].read_data(offset=1, length=2, scaled=False)
+                assert actual_data.dtype == expected_data.dtype
+                compare_arrays(actual_data, expected_data[1:3])
+
+
 def test_lazily_read_channel_data_with_file_path():
     """Test reading channel data lazily after initialising with a file path
     """
@@ -78,19 +122,6 @@ def test_lazily_read_channel_data_with_file_path():
         os.remove(temp_file.name)
 
 
-@pytest.mark.filterwarnings('ignore:.* is deprecated')
-def test_lazily_read_channel_data_with_channel_data_method():
-    """Test reading channel data lazily using the channel_data method of TdmsFile
-    """
-    test_file, expected_data = scenarios.single_segment_with_two_channels().values
-    with test_file.get_tempfile() as temp_file:
-        with TdmsFile.open(temp_file.file) as tdms_file:
-            for ((group, channel), expected_data) in expected_data.items():
-                actual_data = tdms_file.channel_data(group, channel)
-                assert actual_data.dtype == expected_data.dtype
-                np.testing.assert_almost_equal(actual_data, expected_data)
-
-
 @given(offset=strategies.integers(0, 100), length=strategies.integers(0, 100))
 @example(offset=0, length=0)
 @example(offset=0, length=100)
@@ -128,6 +159,7 @@ def test_reading_subset_of_data(offset, length):
             np.testing.assert_equal(channel_subset, expected_data)
 
 
+@settings(suppress_health_check=[HealthCheck.filter_too_much])
 @pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
 @given(offset=strategies.integers(0, 10), length=strategies.integers(0, 10))
 def test_reading_subset_of_data_for_scenario(test_file, expected_data, offset, length):
@@ -259,10 +291,22 @@ def test_indexing_channel_with_integer(index):
                 assert channel_object[index] == expected_channel_data[index]
 
 
-def test_indexing_channel_with_integer_and_caching():
-    """ Test indexing into a channel with an integer index, reusing the same file to test caching
+@given(index=strategies.integers(-8, -1))
+def test_indexing_channel_with_negative_integer(index):
+    """ Test indexing into a channel with a negative integer index
     """
     test_file, expected_data = scenarios.chunked_segment().values
+    with test_file.get_tempfile() as temp_file:
+        with TdmsFile.open(temp_file.file) as tdms_file:
+            for ((group, channel), expected_channel_data) in expected_data.items():
+                channel_object = tdms_file[group][channel]
+                assert channel_object[index] == expected_channel_data[index]
+
+
+@pytest.mark.parametrize("test_file,expected_data", scenarios.get_scenarios())
+def test_indexing_channel_with_integer_and_caching(test_file, expected_data):
+    """ Test indexing into a channel with an integer index, reusing the same file to test caching
+    """
     with test_file.get_tempfile() as temp_file:
         with TdmsFile.open(temp_file.file) as tdms_file:
             for ((group, channel), expected_channel_data) in expected_data.items():
@@ -273,6 +317,18 @@ def test_indexing_channel_with_integer_and_caching():
                 compare_arrays(values, expected_channel_data)
 
 
+@given(index=strategies.integers(0, 1))
+def test_indexing_timestamp_channel_with_integer(index):
+    """ Test indexing into a timestamp data channel with an integer index
+    """
+    test_file, expected_data = scenarios.timestamp_data().values
+    with test_file.get_tempfile() as temp_file:
+        with TdmsFile.open(temp_file.file) as tdms_file:
+            for ((group, channel), expected_channel_data) in expected_data.items():
+                channel_object = tdms_file[group][channel]
+                assert channel_object[index] == expected_channel_data[index]
+
+
 def test_indexing_scaled_channel_with_integer():
     """ Test indexing into a channel with an integer index when the channel is scaled
     """
@@ -400,18 +456,6 @@ def test_read_data_after_close_throws():
         assert "Cannot read data after the underlying TDMS reader is closed" in str(exc_info.value)
 
 
-def test_read_data_after_open_in_read_mode_throws():
-    """ Trying to read channel data after reading all data initially should throw
-    """
-    test_file, expected_data = scenarios.single_segment_with_one_channel().values
-    group, channel = list(expected_data.keys())[0]
-    with test_file.get_tempfile() as temp_file:
-        tdms_file = TdmsFile.read(temp_file.file)
-        with pytest.raises(RuntimeError) as exc_info:
-            tdms_file[group][channel].read_data()
-        assert "Cannot read data after the underlying TDMS reader is closed" in str(exc_info.value)
-
-
 def test_access_data_property_after_opening_throws():
     """ Accessing the data property after opening without reading data should throw
     """
@@ -461,33 +505,63 @@ def test_read_file_passed_as_pathlib_path():
         compare_arrays(channel_obj.data, expected_channel_data)
 
 
-@pytest.mark.filterwarnings('ignore:.* is deprecated')
-def test_get_objects():
-    """Test reading data"""
+def test_read_with_mismatching_index_file():
+    """ Test that reading data when the index file doesn't match the data file raises an error
+    """
 
     test_file = GeneratedFile()
-    test_file.add_segment(*basic_segment())
-    tdms_file = test_file.load()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", 3, 2),
+            channel_metadata("/'group'/'channel2'", 3, 2),
+        ),
+        "01 00 00 00" "02 00 00 00"
+        "03 00 00 00" "04 00 00 00"
+    )
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", 3, 2),
+            channel_metadata("/'group'/'channel2'", 3, 2),
+        ),
+        "01 00 00 00" "02 00 00 00"
+        "03 00 00 00" "04 00 00 00"
+    )
 
-    objects = tdms_file.objects
-    assert len(objects) == 4
-    assert "/" in objects.keys()
-    assert "/'Group'" in objects.keys()
-    assert "/'Group'/'Channel1'" in objects.keys()
-    assert "/'Group'/'Channel2'" in objects.keys()
+    test_file_with_index = GeneratedFile()
+    test_file_with_index.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", 3, 3),
+            channel_metadata("/'group'/'channel2'", 3, 3),
+        ),
+        "01 00 00 00" "02 00 00 00" "03 00 00 00"
+        "04 00 00 00" "05 00 00 00" "06 00 00 00"
+    )
+    test_file_with_index.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", 3, 3),
+            channel_metadata("/'group'/'channel2'", 3, 3),
+        ),
+        "01 00 00 00" "02 00 00 00" "03 00 00 00"
+        "04 00 00 00" "05 00 00 00" "06 00 00 00"
+    )
 
-
-@pytest.mark.filterwarnings('ignore:.* is deprecated')
-def test_get_object_from_group():
-    """Test passing a TdmsGroup to object returns the group"""
-
-    test_file = GeneratedFile()
-    test_file.add_segment(*basic_segment())
-    tdms_file = test_file.load()
-
-    groups = tdms_file.groups()
-    assert tdms_file.object(groups[0]) is groups[0]
-    assert tdms_file.object(groups[0].name) is groups[0]
+    with test_file.get_tempfile(delete=False) as tdms_file:
+        with test_file_with_index.get_tempfile_with_index() as tdms_file_with_index_path:
+            # Move index file from second file to match the name of the first file
+            new_index_file = tdms_file.name + '_index'
+            copyfile(tdms_file_with_index_path + '_index', new_index_file)
+            try:
+                tdms_file.file.close()
+                with pytest.raises(ValueError) as exc_info:
+                    _ = TdmsFile.read(tdms_file.name)
+                assert 'Check that the tdms_index file matches the tdms data file' in str(exc_info.value)
+            finally:
+                os.remove(new_index_file)
+                os.remove(tdms_file.name)
 
 
 def test_get_len_of_file():
@@ -624,6 +698,78 @@ def test_string_data():
         assert expected == read
 
 
+def test_string_data_in_interleaved_segment():
+    """ Test reading a file with string data where a segment
+        has the interleaved data flag set but only one channel.
+        (See issue #239)
+    """
+
+    strings = ["abcdefg", "qwertyuiop"]
+
+    test_file = GeneratedFile()
+    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData")
+    metadata = (
+        "01 00 00 00"  # Number of objects
+        "18 00 00 00")  # Length of the object path
+    metadata += string_hexlify("/'Group'/'StringChannel'")
+    metadata += (
+        "1C 00 00 00"  # Length of index information
+        "20 00 00 00"  # Raw data data type
+        "01 00 00 00"  # Dimension
+        "02 00 00 00 00 00 00 00"  # Number of raw data values
+        "19 00 00 00 00 00 00 00"  # Number of bytes in data
+        "00 00 00 00")  # Number of properties (0)
+    data = (
+        "07 00 00 00"  # index to after first string
+        "11 00 00 00"  # index to after second string
+    )
+    for string in strings:
+        data += string_hexlify(string)
+    test_file.add_segment(toc, metadata, data)
+    tdms_data = test_file.load()
+
+    channel = tdms_data["Group"]["StringChannel"]
+    assert len(channel.data) == len(strings)
+    assert channel.data.dtype == channel.dtype
+    for expected, read in zip(strings, channel.data):
+        assert expected == read
+
+
+def test_incomplete_segment_with_string_data():
+    """ Test incomplete last segment, eg. if LabView crashed, with string data
+    """
+    test_file = GeneratedFile()
+    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
+    metadata = (
+        # Number of objects
+        "01 00 00 00"
+        # Length of the object path
+        "18 00 00 00")
+    metadata += string_hexlify("/'Group'/'StringChannel'")
+    metadata += (
+        # Length of index information
+        "1C 00 00 00"
+        # Raw data data type
+        "20 00 00 00"
+        # Dimension
+        "01 00 00 00"
+        # Number of raw data values
+        "02 00 00 00"
+        "00 00 00 00"
+        # Number of bytes in data
+        "19 00 00 00"
+        "00 00 00 00"
+        # Number of properties (0)
+        "00 00 00 00")
+    data = "00 00 00 00"
+    test_file.add_segment(toc, metadata, data, incomplete=True)
+
+    tdms_data = test_file.load()
+
+    channel = tdms_data["Group"]["StringChannel"]
+    assert len(channel) == 0
+
+
 def test_slash_and_space_in_name():
     """Test name like '01/02/03 something'"""
 
@@ -675,44 +821,27 @@ def test_single_quote_in_name():
     assert len(data_1) == 2
 
 
-@pytest.mark.filterwarnings('ignore:.* is deprecated')
-def test_root_object_paths():
-    """Test the group and channel properties for the root object"""
-    test_file = GeneratedFile()
-    test_file.add_segment(*basic_segment())
-    tdms_data = test_file.load()
-
-    obj = tdms_data.object()
-    assert obj.group is None
-    assert obj.channel is None
-
-
-@pytest.mark.filterwarnings('ignore:.* is deprecated')
 def test_group_object_paths():
-    """Test the group and channel properties for a group"""
+    """Test the path and name properties for a group"""
     test_file = GeneratedFile()
     test_file.add_segment(*basic_segment())
     tdms_data = test_file.load()
 
-    obj = tdms_data["Group"]
-    assert obj.path == "/'Group'"
-    assert obj.name == "Group"
-    assert obj.group == "Group"
-    assert obj.channel is None
+    group = tdms_data["Group"]
+    assert group.path == "/'Group'"
+    assert group.name == "Group"
 
 
-@pytest.mark.filterwarnings('ignore:.* is deprecated')
 def test_channel_object_paths():
-    """Test the group and channel properties for a group"""
+    """Test the path and name properties for a channel"""
     test_file = GeneratedFile()
     test_file.add_segment(*basic_segment())
     tdms_data = test_file.load()
 
-    obj = tdms_data["Group"]["Channel1"]
-    assert obj.path == "/'Group'/'Channel1'"
-    assert obj.name == "Channel1"
-    assert obj.group == "Group"
-    assert obj.channel == "Channel1"
+    channel = tdms_data["Group"]["Channel1"]
+    assert channel.path == "/'Group'/'Channel1'"
+    assert channel.name == "Channel1"
+    assert channel.group_name == "Group"
 
 
 def test_object_repr():
@@ -768,3 +897,142 @@ def test_debug_logging(caplog):
 
     assert "Reading metadata for object /'group'/'channel1' with index header 0x00000014" in caplog.text
     assert "Object data type: Int32" in caplog.text
+
+
+def test_memory_released_when_tdms_file_out_of_scope():
+    """ Tests that when a TDMS file object goes out of scope,
+        TDMS channels and their data are also freed.
+        This ensures there are no circular references between a TDMS file
+        and its channels, which would mean the GC is needed to free these objects.
+    """
+
+    test_file, expected_data = scenarios.single_segment_with_one_channel().values
+    with test_file.get_tempfile() as temp_file:
+        tdms_data = TdmsFile.read(temp_file.file)
+        chan = tdms_data['group']['channel1']
+        chan_ref = weakref.ref(chan)
+        data_ref = weakref.ref(chan.data)
+        raw_data_ref = weakref.ref(chan.raw_data)
+    del tdms_data
+    del chan
+
+    assert raw_data_ref() is None
+    assert data_ref() is None
+    assert chan_ref() is None
+
+
+def test_close_after_read():
+    test_file, _ = scenarios.single_segment_with_one_channel().values
+    temp_file = test_file.get_tempfile(delete=False)
+    try:
+        temp_file.file.close()
+        tdms_data = TdmsFile.read(temp_file.name)
+        tdms_data.close()
+    finally:
+        os.remove(temp_file.name)
+
+
+def test_multiple_close_after_open():
+    test_file, _ = scenarios.single_segment_with_one_channel().values
+    temp_file = test_file.get_tempfile(delete=False)
+    try:
+        temp_file.file.close()
+        with TdmsFile.open(temp_file.name) as tdms_data:
+            tdms_data.close()
+        tdms_data.close()
+    finally:
+        os.remove(temp_file.name)
+
+
+def test_interleaved_segment_different_length():
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", 3, 3),
+            channel_metadata("/'group'/'channel2'", 3, 2),
+        ),
+        "01 00 00 00" "02 00 00 00"
+        "01 00 00 00" "02 00 00 00"
+        "01 00 00 00"
+    )
+
+    with pytest.raises(ValueError) as exc_info:
+        _ = test_file.load()
+    assert str(exc_info.value) == "Cannot read interleaved data with different chunk sizes"
+
+
+def test_multiple_string_channels_with_interleaved_flag():
+    """ Test reading a segment with the interleaved data flag set but multiple string channels throws an exception
+    """
+
+    test_file = GeneratedFile()
+    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData")
+    metadata = "02 00 00 00"  # Number of objects
+    for channel in ["Channel1", "Channel2"]:
+        channel_path = "/'Group'/'%s'" % channel
+        metadata += hexlify_value('<L', len(channel_path))
+        metadata += string_hexlify(channel_path)
+        metadata += (
+            "1C 00 00 00"  # Length of index information
+            "20 00 00 00"  # Raw data data type
+            "01 00 00 00"  # Dimension
+            "02 00 00 00 00 00 00 00"  # Number of raw data values
+            "02 00 00 00 00 00 00 00"  # Number of bytes in data
+            "00 00 00 00")  # Number of properties
+    data = "01 00 00 00 02 00 00 00 03 00 00 00 04 00 00 00"
+    test_file.add_segment(toc, metadata, data)
+    with pytest.raises(ValueError) as exc_info:
+        _ = test_file.load()
+    assert str(exc_info.value) == "Cannot read interleaved segment containing channels with unsized types"
+
+
+def test_interleaved_segment_with_sized_and_unsized_types():
+    test_file = GeneratedFile()
+    string_channel_metadata = (
+        "18 00 00 00" +  # Length of the object path
+        string_hexlify("/'group'/'StringChannel'") +
+        "1C 00 00 00"  # Length of index information
+        "20 00 00 00"  # Raw data data type
+        "01 00 00 00"  # Dimension
+        "02 00 00 00 00 00 00 00"  # Number of raw data values
+        "02 00 00 00 00 00 00 00"  # Number of bytes in data
+        "00 00 00 00")  # Number of properties (0)
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", 3, 2),
+            string_channel_metadata,
+        ),
+        "01 00 00 00" "02 00 00 00"
+        "01 00 00 00" "02 00 00 00"
+    )
+
+    with pytest.raises(ValueError) as exc_info:
+        _ = test_file.load()
+    assert str(exc_info.value) == "Cannot read interleaved segment containing channels with unsized types"
+
+
+def test_no_new_obj_list_for_first_segment():
+    """ Test when kTocNewObjList is not set in the first segment
+    """
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", 3, 4),
+        ),
+        "01 00 00 00" "02 00 00 00" "03 00 00 00" "04 00 00 00"
+    )
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", 3, 4),
+        ),
+        "05 00 00 00" "06 00 00 00" "07 00 00 00" "08 00 00 00"
+    )
+
+    with test_file.get_tempfile() as temp_file:
+        tdms_data = TdmsFile.read(temp_file.file)
+    channel_data = tdms_data['group']['channel1']
+    compare_arrays(channel_data[:], [1, 2, 3, 4, 5, 6, 7, 8])
