@@ -8,7 +8,7 @@ except ImportError:
     OrderedDict = dict
 import numpy as np
 
-from nptdms.writer import TdmsSegment, read_properties_dict
+from nptdms.writer import ChannelObject, TdmsSegment, read_properties_dict
 from nptdms.types import *
 
 
@@ -176,6 +176,7 @@ def test_writing_properties_with_numpy_typed_values():
         "int64prop": np.int64(64),
         "float32prop": np.float32(32.0),
         "float64prop": np.float64(64.0),
+        "boolprop": np.bool_(True),
     }
 
     tdms_properties = read_properties_dict(properties)
@@ -185,6 +186,7 @@ def test_writing_properties_with_numpy_typed_values():
     assert tdms_properties["int64prop"] == Int64(64)
     assert tdms_properties["float32prop"] == SingleFloat(32.0)
     assert tdms_properties["float64prop"] == DoubleFloat(64.0)
+    assert tdms_properties["boolprop"] == Boolean(True)
 
 
 def test_error_raised_when_cannot_convert_property_value():
@@ -196,6 +198,20 @@ def test_error_raised_when_cannot_convert_property_value():
         read_properties_dict(properties)
 
 
+@pytest.mark.parametrize(
+    "data",
+    [
+        np.array(0.0, dtype=np.dtype('float32')),
+        np.zeros((2, 2), dtype=np.dtype('float32')),
+        np.zeros((2, 2, 2), dtype=np.dtype('float32')),
+    ]
+)
+def test_channel_with_invalid_data(data):
+    with pytest.raises(ValueError) as exc_info:
+        _ = ChannelObject("Group", "Channel", data, {})
+    assert str(exc_info.value) == "Channel data must be a 1d array"
+
+
 def _assert_sequence_equal(values, expected_values):
     position = 1
     expected_values = iter(expected_values)
