@@ -2,6 +2,7 @@
 
 import numpy as np
 import struct
+from nptdms.timestamp import TdmsTimestamp, TimestampArray
 
 
 __all__ = [
@@ -41,13 +42,13 @@ tds_data_types = {}
 numpy_data_types = {}
 
 
-def tds_data_type(enum_value, np_type):
+def tds_data_type(enum_value, np_type, set_np_type=True):
     def decorator(cls):
         cls.enum_value = enum_value
         cls.nptype = None if np_type is None else np.dtype(np_type)
         if enum_value is not None:
             tds_data_types[enum_value] = cls
-        if np_type is not None:
+        if set_np_type and np_type is not None:
             numpy_data_types[np.dtype(np_type)] = cls
         return cls
     return decorator
@@ -85,6 +86,7 @@ class Bytes(TdmsType):
 
 class StructType(TdmsType):
     struct_declaration = None
+    nptype = None
 
     def __init__(self, value):
         self.value = value
@@ -95,6 +97,14 @@ class StructType(TdmsType):
         read_bytes = file.read(cls.size)
         return _struct_unpack(endianness + cls.struct_declaration, read_bytes)[0]
 
+    @classmethod
+    def from_bytes(cls, byte_array, endianness="<"):
+        """ Convert an array of bytes into a numpy array of data
+        """
+        array = byte_array.view()
+        array.dtype = cls.nptype.newbyteorder(endianness)
+        return array
+
 
 @tds_data_type(0, None)
 class Void(TdmsType):
@@ -166,13 +176,13 @@ class ExtendedFloat(TdmsType):
     pass
 
 
-@tds_data_type(0x19, np.single)
+@tds_data_type(0x19, np.single, set_np_type=False)
 class SingleFloatWithUnit(StructType):
     size = 4
     struct_declaration = "f"
 
 
-@tds_data_type(0x1A, np.double)
+@tds_data_type(0x1A, np.double, set_np_type=False)
 class DoubleFloatWithUnit(StructType):
     size = 8
     struct_declaration = "d"
@@ -219,6 +229,10 @@ class Boolean(StructType):
     size = 1
     struct_declaration = "b"
 
+    def __init__(self, value):
+        self.value = bool(value)
+        self.bytes = _struct_pack('<' + self.struct_declaration, self.value)
+
     @classmethod
     def read(cls, file, endianness="<"):
         return bool(super(Boolean, cls).read(file, endianness))
@@ -230,9 +244,6 @@ class TimeStamp(TdmsType):
     # 01/01/1904 00:00:00.00 UTC, ignoring leap seconds,
     # and number of 2^-64 fractions of a second.
     # Note that the TDMS epoch is not the Unix epoch.
-
-    # We convert times to numpy datetime64s with microsecond precision,
-    # so lose some precision compared with  TDMS.
     _tdms_epoch = np.datetime64('1904-01-01 00:00:00', 'us')
     _fractions_per_microsecond = float(10**-6) / 2**-64
 
@@ -263,11 +274,7 @@ class TimeStamp(TdmsType):
         else:
             (seconds, second_fractions) = _struct_unpack(
                  endianness + 'qQ', data)
-        micro_seconds = int(
-            float(second_fractions) / cls._fractions_per_microsecond)
-
-        return (cls._tdms_epoch + np.timedelta64(seconds, 's') +
-                np.timedelta64(micro_seconds, 'us'))
+        return TdmsTimestamp(seconds, second_fractions)
 
     @classmethod
     def from_bytes(cls, byte_array, endianness="<"):
@@ -275,15 +282,10 @@ class TimeStamp(TdmsType):
         """
         byte_array = byte_array.reshape((-1, 16))
         if endianness == "<":
-            second_fractions = byte_array[:, 0:8].ravel()
-            seconds = byte_array[:, 8:16].ravel()
+            dtype = np.dtype([('second_fractions', '<u8'), ('seconds', '<i8')])
         else:
-            seconds = byte_array[:, 0:8].ravel()
-            second_fractions = byte_array[:, 8:16].ravel()
-        seconds.dtype = np.dtype('timedelta64[s]').newbyteorder(endianness)
-        second_fractions.dtype = np.dtype('uint64').newbyteorder(endianness)
-        micro_seconds = (second_fractions / cls._fractions_per_microsecond) * np.timedelta64(1, 'us')
-        return cls._tdms_epoch + seconds + micro_seconds
+            dtype = np.dtype([('seconds', '>i8'), ('second_fractions', '>u8')])
+        return TimestampArray(byte_array.view(dtype).reshape(-1))
 
 
 @tds_data_type(0x08000c, np.complex64)
