@@ -229,6 +229,8 @@ class ChannelObject(TdmsObject):
         self.group = group
         self.channel = channel
         self.data = _to_np_array(data)
+        if self.data.ndim != 1:
+            raise ValueError("Channel data must be a 1d array")
         self.properties = properties
 
     @property
@@ -266,7 +268,7 @@ def _to_tdms_value(value):
         return numpy_data_types[value.dtype](value)
     if isinstance(value, TdmsType):
         return value
-    if isinstance(value, bool):
+    if isinstance(value, bool) or isinstance(value, np.bool_):
         return Boolean(value)
     if isinstance(value, (int, long)):
         return to_int_property_value(value)
@@ -316,8 +318,7 @@ def to_file(file, array):
     try:
         array.tofile(file)
     except (TypeError, IOError, UnsupportedOperation):
-        # tostring actually returns bytes
-        file.write(array.tostring())
+        file.write(array.tobytes())
 
 
 def write_values(file, array):
